/*
 * Copyright (C) 2001 Igor Sysoev.
 */

#include "mod_accel.h"

/*
   Cache file:
        Date LastModifed Expires Flags Length LastAccessed
        URL
        HTTP status
        Header
        CRLF
        Body
*/

char *cache_state_line_mask = CACHE_STATE_LINE_MASK;

int accel_check_cachable(accel_rec *a)
{
    if (a->dir_conf->no_cache                   /* AccelNoCache              */
        || a->r->method_number != M_GET)        /* check cache for GET       */
    {                                           /* and HEAD requests only    */
        a->accel_state = ACCEL_PASS;
        return FALSE;
    }

    /* don't cache request with "accel_nocache" note or env var */

    if (ap_table_get(a->r->subprocess_env, "ACCEL_NOCACHE")
        || ap_table_get(a->r->notes, "accel_nocache"))
    {
        a->accel_state = ACCEL_NTNC;
        a->backend_cache = ACCEL_NNC;
        a->note_nocache = 1;
        a->expired = 0;
        return FALSE;
    }

    /* don't cache request with "Authorization" header
       if user should not be revalidated
       or "Authorization" header should not be ignored */

    if (!a->dir_conf->ignore_auth               /* AccelIgnoreAuth           */
        && !a->dir_conf->revalidate_user        /* AccelRevalidateUser       */
        && ap_table_get(a->r->headers_in, "Authorization") != NULL)
    {
        a->accel_state = ACCEL_AUTH;
        return FALSE;
    }

    return TRUE;
}

/*
   Return TRUE if cached response is valid
          and FALSE if there's no cached response or it's must be revalidated
*/

int accel_check_cache(accel_rec *a)
{
    int          max_age;
    const char *cache_control;

    int cache = ACCEL_BL_NONE;
    request_rec *r = a->r;
    accel_dir_conf *dcf = a->dir_conf;

    accel_make_cachefile_name(a, 1);

    if (a->invalidate) {
        a->accel_state = ACCEL_INVL;
        return FALSE;
    }

    if (!dcf->ignore_no_cache                   /* AccelIgnoreNoCache        */
        && (ap_find_token(r->pool,
                          ap_table_get(r->headers_in, "Pragma"), "no-cache"))) {
        a->invalidate = 1;
        a->accel_state = ACCEL_PGNC;
        return FALSE;
    }

    cache_control = ap_table_get(r->headers_in, "Cache-Control");
    max_age = accel_max_age(r->pool, cache_control);

    if (!dcf->ignore_no_cache                   /* AccelIgnoreNoCache        */
        && (max_age == 0
            || ap_find_token(r->pool, cache_control, "no-cache")))
    {
        a->invalidate = 1;
        a->accel_state = ACCEL_PGNC;
        return FALSE;
    }

    a->accel_state = ACCEL_MISS;

    /* open existent cache file */
    if ((a->cache_fd = ap_popenf(r->pool, a->cachefile,
                                 O_RDWR | O_BINARY, 0640)) >= 0)
    {
        if (!accel_read_cachefile_header(a))
            return FALSE;

    /* file not found, look - may be someone is already getting it ? */
    } else if (accel_open_cachefile_with_busy_lock(a, dcf->busy_lock)
               != ACCEL_BL_OK) {
        return FALSE;
    }

    /* Cache-Control: max-age=# */
    if (!dcf->ignore_no_cache
        && max_age > 0
        && a->time - a->cache_date > max_age)
    {
        if (is_max_connected(a)
            || (cache = accel_open_cachefile_with_busy_lock(a, dcf->busy_lock))
                      != ACCEL_BL_OK) {

            if (cache == ACCEL_BL_OLD)
                a->revalidate = 1;

            a->invalidate = 1;
            a->accel_state = ACCEL_AGED;
            return FALSE;
        }
    }

    a->expired_time = a->time - a->cache_expires;

    if (dcf->revalidate_user) {                     /* AccelRevalidateHeader */
        a->accel_state = ACCEL_RVUS;
        return FALSE;
    }

    a->accel_state = ACCEL_EXPR;

#ifdef MOD_ACCEL_DEBUG
    ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, r,
                  "accel: expired time: %d", a->expired_time);
#endif

    /* not expired or max waiting */
    if (a->expired_time < 0 || is_max_connected(a)) {
        a->accel_state = ACCEL_HIT;
        return TRUE;
    }

    /* expired but less then AccelMaxStale */
    if (a->expired_time > 0 && a->expired_time <= dcf->max_stale) {
        int busy;

        /* check busy with expired busy lock and no wait state
           if busy or max waiting then HIT */
        if (busy = accel_check_busy_lock(a, dcf->expired_busy_lock, 0)
            || is_max_connected(a))
        {
            if (busy)
                a->lock_state |= ACCEL_UPDATING;

            a->accel_state = ACCEL_HIT;
            a->lock_time = 0;
            return TRUE;
        }

        a->lock_time = 0;
        a->revalidate = 1;
        return FALSE;
    }

    /* expired, open file with non-expired busy lock */
    if ((cache = accel_open_cachefile_with_busy_lock(a, dcf->busy_lock))
               != ACCEL_BL_OK) {

        if (cache == ACCEL_BL_OLD)
            a->revalidate = 1;

        return FALSE;
    }

    a->expired_time = a->time - a->cache_expires;

#ifdef MOD_ACCEL_DEBUG
    ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, r,
                  "accel: 2nd expired time: %d", a->expired_time);
#endif

    /* not expired or max waiting */
    if (a->expired_time < 0 || is_max_connected(a)) {
        a->accel_state = ACCEL_HIT;
        return TRUE;
    }

    a->revalidate = 1;
    return FALSE;
}

int accel_open_cachefile_with_busy_lock(accel_rec *a, int busy_lock)
{
    int          busy;
    struct stat  sb;

    request_rec *r = a->r;
    ino_t inode = 0;

    if (busy_lock <= 0)
        return a->cache_fd >= 0 ? ACCEL_BL_OLD : ACCEL_BL_NONE;

    if (a->cache_fd >= 0) {
        if (fstat(a->cache_fd, &sb) == -1)
            return ACCEL_BL_ERR;
        inode = sb.st_ino;
    }

    if (a->lock_time == -1)
        a->lock_time = 0;

    while (a->lock_time < a->dir_conf->max_busy_lock) {
#ifdef MOD_ACCEL_DEBUG
        ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, a->r,
                      "accel: busy lock %d", a->lock_time);
#endif
        busy = accel_check_busy_lock(a, busy_lock,
                                     a->dir_conf->max_busy_lock - a->lock_time);

#ifdef MOD_ACCEL_DEBUG
        ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, a->r,
                      "accel: conn:wait %d:%d",
                      a->max_connected, a->max_waiting);
#endif

        /* if request takes time more then busy lock time then
           there are can be several processes make the same requests
           and one of them can be complete at this time */

        if (inode) {                        /* we have expired cache file */

            if (a->lock_time == 0) {        /* first busy lock interation ? */
                if (is_max_connected(a))
                    return ACCEL_BL_OK;

            } else if (stat(a->cachefile, &sb) != -1) {

                if (sb.st_ino == inode) {   /* cache file is old */
                    if (is_max_connected(a))
                        return ACCEL_BL_OK;

                } else {                    /* cache file updated */
                    int fd;

                    if ((fd = ap_popenf(r->pool, a->cachefile,
                                        O_RDWR | O_BINARY, 0640)) >= 0) {

                        ap_pclosef(r->pool, a->cache_fd);
                        a->cache_fd = fd;
                        a->header = BACKEND;
                        a->body = BACKEND;
                        a->lock_state |= ACCEL_NEW_HIT;

                        return accel_read_cachefile_header(a) ? ACCEL_BL_OK:
                                                                ACCEL_BL_ERR;

                    } else {
                        return is_max_connected(a) ? ACCEL_BL_OK : ACCEL_BL_ERR;
                    }
                }

            } else {                        /* can't stat cache file */
                if (is_max_connected(a))
                    return ACCEL_BL_OK;

                if (errno != ENOENT)
                    return ACCEL_BL_ERR;
            }

        } else {                            /* no cache file at all */
            if (a->lock_time > 0
                && (a->cache_fd = ap_popenf(r->pool, a->cachefile,
                                         O_RDWR | O_BINARY, 0640)) >= 0)

                return accel_read_cachefile_header(a) ? ACCEL_BL_OK:
                                                        ACCEL_BL_ERR;

            if (a->max_waiting)
                return ACCEL_BL_NONE;
        }

        if (!busy)
            return inode ? ACCEL_BL_OLD : ACCEL_BL_NONE;

        sleep(1);
        a->time++;
        a->lock_time++;
    }

    return inode ? ACCEL_BL_OLD : ACCEL_BL_NONE;
}

int accel_read_cachefile_header(accel_rec *a)
{
    char buf[HUGE_STRING_LEN];
    request_rec *r = a->r;

    if (accel_gets(buf, HUGE_STRING_LEN, a, CACHE) <= 0
        || !ap_checkmask(buf, cache_state_line_mask))
    {
        return FALSE;
    }

#ifdef MOD_ACCEL_DEBUG
    ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, r,
                  "accel: state line '%s'", buf);
#endif

    sscanf(buf, CACHE_STATE_READ_FORMAT, &a->cache_date, &a->cache_lm,
           &a->cache_expires, &a->cache_flags, &a->cache_len);

    /* URL */
    if (accel_gets(buf, HUGE_STRING_LEN, a, CACHE) <= 0)
        return FALSE;

#ifdef MOD_ACCEL_API
    a->cnt->in_flags = a->cnt->out_flags = a->cache_flags;
#endif

    if (!accel_read_http_header(a, CACHE))
        return FALSE;

    a->header = CACHE;
    a->body = CACHE;

#ifdef MOD_ACCEL_DEBUG
    accel_log_header(r, "cache", a->cache_header->line,
                     a->cache_header->headers_out);
#endif

    return TRUE;
}

void accel_send_cache_range(accel_rec *a)
{
    long off, len;

    request_rec *r = a->r;
    FILE *f = ap_pfdopen(r->pool, a->cache_fd, "r");

    while (ap_each_byterange(r, &off, &len)) {
        if (a->cache_header_size + off < a->cache_size) {
            int size, rest;

            if (a->cache_size < a->cache_header_size + off + len) {
                size = a->cache_size - (a->cache_header_size + off);
                rest = len - size;

            } else {
                size = len;
                rest = 0;
            }

            ap_soft_timeout("accel: send range to client", r);
            ap_rwrite(a->cache_buf + a->cache_header_size + off, size, r);
            ap_reset_timeout(r);

            if (rest) {
                if (a->cache_size != a->cache_seek) {
                    if (fseek(f, a->cache_size, SEEK_SET)) {
                        ap_log_rerror(APLOG_MARK, APLOG_ERR, r,
                                      "Failed to fseek for byterange "
                                      "(%ld, %ld)", off, len);
                        a->cache_error = 1;
                        return;
                    }
                }

                ap_send_fd_length(f, r, rest);
                a->cache_seek = a->cache_size + rest;
            }

        } else {
            if (fseek(f, a->cache_header_size + off, SEEK_SET)) {
                ap_log_rerror(APLOG_MARK, APLOG_ERR, r,
                              "Failed to fseek for byterange (%ld, %ld)",
                              off, len);
                a->cache_error = 1;
                return;

            } else {
                ap_send_fd_length(f, r, len);
                a->cache_seek = a->cache_header_size + off + len;
            }
        }
    }
}

void accel_send_cache_body(accel_rec *a)
{
    int          rsize, wsize, first, tail, osize, header, pos;

    request_rec *r = a->r;

    first = 1;

    header = a->cache_header_size;
    pos = 0;
    osize = rsize = a->cache_size;

    ap_soft_timeout("accel: send body to client", r);

    while (!r->connection->aborted) {

#ifdef MOD_ACCEL_DEBUG
        ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, r,
                      "accel: reading %d bytes from cache", rsize);

        if (rsize == 0) {
            ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, r,
                          "accel: len %d:%d", osize, a->cache_len);
            break;
        }
#endif

        if (osize < a->cache_len) {
             wsize = rsize - header - accel_content_tail + pos;
             tail = TRUE;

        } else {
             wsize = rsize - header + pos;
             tail = FALSE;
        }

#ifdef MOD_ACCEL_DEBUG
        ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, r,
            "accel: writing %d:%d:%d bytes to client", rsize, wsize, tail);
#endif

#ifdef MOD_ACCEL_API
        a->cnt->buf = a->cache_buf + header;
        a->cnt->len = wsize;
        a->cnt->state = 0;

        if (first) {
            a->cnt->state |= CNT_FIRST;
            first = 0;
        }

        if (tail)
            a->cnt->state |= CNT_TAIL;

        ap_hook_call("mod_accel::content", r, a->cnt);
#endif
        ap_rwrite(a->cache_buf + header, wsize, r);
        ap_reset_timeout(r);

        if (osize >= a->cache_len) {
#ifdef MOD_ACCEL_DEBUG
            ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, r,
                          "accel: len %d:%d", osize, a->cache_len);
#endif
            break;
        }

        if (tail) {
            pos = accel_content_tail;
            memcpy(a->cache_buf, a->cache_buf + header + wsize, pos);
        }

        header = 0;

        if ((rsize = read(a->cache_fd, a->cache_buf + pos,
                          a->send_threshold - pos))
                   == -1)
        {
            ap_log_rerror(APLOG_MARK, APLOG_ERR, r,
                          "accel: error reading cache file %s", a->cachefile);
            a->cache_error = 1;
        }

        osize += rsize;
    }

    ap_kill_timeout(r);

    accel_update_cache(a);
}

/*
   Return TRUE if HTTP_NOT_MODIFIED status can be returned to client
*/

int accel_check_ims(accel_rec *a, int last_modified, header_rec *h)
{
    int rc;

    if (!last_modified)
        return OK;

    /* return OK if cached content should be changed */

#ifdef MOD_ACCEL_API
    if (((a->cache_flags & (CNT_NM_CHECKED | CNT_NM_DISABLED))
                        == (CNT_NM_CHECKED | CNT_NM_DISABLED)))
        return OK;
#endif

    /* return HTTP_NOT_MODIFIED if client asks IMS and LM <= IMS
       no hooks or no one hook returns FALSE */

    a->r->mtime = last_modified;
    ap_table_setn(a->r->headers_out, "ETag",
                  ap_table_get(h->headers_out, "ETag"));

    if ((rc = ap_meets_conditions(a->r)) == OK || rc != HTTP_NOT_MODIFIED)
        return rc;

#ifdef MOD_ACCEL_API
    if (((a->cache_flags & (CNT_NM_CHECKED | CNT_NM_DISABLED))
                         == CNT_NM_CHECKED))
        return HTTP_NOT_MODIFIED;

    /* return OK if any hook returns DECLINED */
    if (ap_hook_call("mod_accel::ims", &rc, a->r, h))
        return OK;
#endif

    return HTTP_NOT_MODIFIED;
}

void accel_update_cache(accel_rec *a)
{
    struct stat  sb;
    request_rec *r = a->r;

#ifdef MOD_ACCEL_DEBUG
    ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, a->r,
                  "accel: update cache: %d", a->expired);
#endif

    if (a->note_nocache)
        return;
 
    if (a->cachefile && stat(a->cachefile, &sb) == 0) {
        a->cachefile_exists = 1;
        a->cachefile_path_exists = 1;
        if (a->cachefile_existed == -1)
            a->cachefile_existed = 1;
    }

    if (!is_cachable(a) || a->expired || a->backend_aborted) {
        accel_unlink_cachefile(a);
        return;
    }

    if (a->dir_conf->revalidate_user) {
        if (a->backend_header->status == HTTP_UNAUTHORIZED
            || a->backend_header->status == HTTP_MOVED_TEMPORARILY)
        {
            return;
        }

        if (a->backend_header->status != HTTP_OK
            && a->backend_header->status != HTTP_NOT_MODIFIED)
        {
            accel_unlink_cachefile(a);
            return;
        }

    } else if (a->header == BACKEND
               && a->backend_header->status != HTTP_OK
               && a->backend_header->status != HTTP_MOVED_PERMANENTLY
               && a->backend_header->status != HTTP_MOVED_TEMPORARILY) {
        accel_unlink_cachefile(a);
        return;
    }

    a->time = time(NULL);

    if (!a->cache_lm)
        a->cache_lm = a->time;

    if (!a->cache_len && a->header_end)
        a->cache_len = a->header_end + a->incnt;

    a->cache_state_line = ap_psprintf(r->pool, CACHE_STATE_WRITE_FORMAT,
                                      a->cache_date, a->cache_lm,
                                      a->cache_expires, a->cache_flags,
                                      a->cache_len, a->time);

    if (a->body == BACKEND) {
        if (a->temp_fd < 0) {

            if (!accel_open_tempfile(a, ACCEL_NO_STUB)
                || !accel_tempfile_write(a, a->buf, a->incnt,
                                        (off_t) a->header_end)) {
                accel_unlink_cachefile(a);
                return;
            }

        } else {
            if (!accel_tempfile_write(a, a->cache_state_line,
                                      CACHE_STATE_LINE_SIZE, (off_t) 0))
            {
                accel_unlink_cachefile(a);
                return;
            }
        }

    } else {
        if (!accel_update_cachefile(a)) {
            accel_unlink_cachefile(a);
            return;
        }
    }

    if (a->tempfile) {
        if (!a->cachefile_path_exists && a->server_conf->auto_create_path)
            accel_create_cache_path(a);

        accel_unlink_cachefile(a);
        if (link(a->tempfile, a->cachefile) == -1) {
            ap_log_rerror(APLOG_MARK, APLOG_ERR, r,
                          "accel: error linking cache file %s to %s",
                          a->tempfile, a->cachefile);
            a->cache_error = 1;
        } else {
            a->cachefile_exists = 1;
            a->cache_fd = a->temp_fd;
        }
    }

    accel_clear_busy_lock(a);
}


void accel_response_expired(accel_rec *a)
{
    int         age, date, expires;
    char       *token;
    const char *accel_expires, *cache_control;

    int status = 0;
    table *headers = NULL;
    request_rec *r = a->r;

    if (a->expired != -1)
        return;

    if (a->backend_aborted) {
        a->expired = 1;
        return;
    }

    if (a->dir_conf->revalidate_user
        && (a->backend_header->status == HTTP_UNAUTHORIZED
            || a->backend_header->status == HTTP_MOVED_TEMPORARILY))
    {
        a->expired = 0;
        return;
    }

    if (a->header == BACKEND) {
        status = a->backend_header->status;
        headers = a->backend_header->headers_out;

    } else if (a->revalidate || a->dir_conf->revalidate_user) {
        status = a->cache_header->status;
        headers = a->cache_header->headers_out;
    }

    if (status != HTTP_OK
        && status != HTTP_MOVED_PERMANENTLY
        && status != HTTP_MOVED_TEMPORARILY
        && status != HTTP_NOT_MODIFIED)
    {
        a->expired = 1;
        return;
    }

    a->cache_date = a->time;

    date = ap_parseHTTPdate(ap_table_get(headers, "Date"));
    a->cache_expires = 0;
    a->cache_lm = ap_parseHTTPdate(ap_table_get(headers, "Last-Modified"));

    if ((accel_expires = ap_table_get(headers, "X-Accel-Expires"))
         && (age = atoi(accel_expires)) >= 0)
    {
        a->cache_expires = date + age;
        a->backend_cache = ACCEL_XAE;
        a->backend_cache_time = age;

        a->expired = a->cache_expires <= date;
        return;
    }

    if (!a->dir_conf->ignore_expires) {

        if (cache_control = ap_table_get(headers, "Cache-Control")) {

            if (ap_find_token(r->pool, cache_control, "no-cache")) {
                a->cache_expires = date;
                a->backend_cache = ACCEL_CTL;
                a->backend_cache_time = ACCEL_NO_CACHE;

                a->expired = 1;
                return;

            } else if ((age = accel_max_age(r->pool, cache_control)) >= 0) {
                a->cache_expires = date + age;
                a->backend_cache = ACCEL_CTL;
                a->backend_cache_time = age;

                a->expired = a->cache_expires <= date;
                return;
            }
        }

        if (expires = ap_parseHTTPdate(ap_table_get(headers, "Expires"))) {
            a->cache_expires = expires;
            a->backend_cache = ACCEL_EXP;
            a->backend_cache_time = expires - date;

            a->expired = a->cache_expires <= date;
            return;
        }
    }

    /* cache 301 for (AccelCleanLastAccessed - 1) seconds */
    if (a->backend_header->status == HTTP_MOVED_PERMANENTLY) {
        a->cache_expires = a->time + a->server_conf->clean_la;
        a->backend_cache = ACCEL_MVD;
        a->backend_cache_time = a->server_conf->clean_la;

        a->expired = 0;
        return;
    }

    /* cache 302 if there're X-Accel-Expires, Cache-Control or Expires */
    if (a->backend_header->status == HTTP_MOVED_TEMPORARILY) {
        a->expired = 1;
        return;
    }

    if (a->cache_lm && a->dir_conf->cache_lm_factor > 0) {
        a->cache_expires = a->time
                + (((int64_t) (date - a->cache_lm))
                                         * a->dir_conf->cache_lm_factor) / 100;

        a->backend_cache = ACCEL_LMF;
        a->backend_cache_time = a->cache_expires - a->time;

        a->expired = a->cache_expires <= a->time;
        return;
    }

    if (a->dir_conf->default_expire > 0) {
        a->cache_expires = a->time + a->dir_conf->default_expire;
        a->backend_cache = ACCEL_ADE;
        a->backend_cache_time = a->dir_conf->default_expire;

        a->expired = a->cache_expires <= a->time;
        return;
    }

    a->expired = 1;
    return;
}

int accel_compare_cookie(const void *one, const void *two)
{
    return strcmp(*(char **) one, *(char **) two);
}

void accel_make_cachefile_name(accel_rec *a, int cache_cookie)
{
    int         i, j, n, m;
    const char *host = NULL, *ptr, *pair, *name, *value;
    char        filename[MD5_LEN + 4 + 1]; /* 4 is for 4 slashes */

    request_rec *r = a->r;
    int *level = (int *) a->server_conf->levels->elts;

    a->key = a->url;

    if (a->pass->the_same_host)
        host = r->connection->local_ip;
    else if (a->pass->preserve_host)
        host = ap_table_get(r->headers_in, "Host");

    if (host) {
        char *port, *prefix, *rest;

        if (port = strchr(host, ':'))
            host = ap_pstrndup(r->pool, host, port - host);

        prefix = strstr(a->url, "://");
        prefix += 3;

        if (!(rest = strchr(prefix, ':')))
            rest = strchr(prefix, '/');

        a->key = ap_pstrcat(r->pool, ap_pstrndup(r->pool,
                                                 a->url, prefix - a->url),
                                     host, rest, NULL);
    }

    if (cache_cookie
        && a->dir_conf->cache_cookie
        && (ptr = ap_table_get(r->headers_in, "Cookie")) != NULL)
    {
        array_header *cookies = ap_make_array(r->pool,
                                              a->dir_conf->cache_cookies->nelts,
                                              sizeof(char *));
        accel_cookie *cookie =
                     (accel_cookie *) a->dir_conf->cache_cookies->elts;
        accel_cookie *no_cookie =
                     (accel_cookie *) a->dir_conf->no_cache_cookies->elts;

        while (*ptr && (pair = ap_get_token(r->pool, &ptr, 0))) {
            int found = 0; 

            if (*ptr == ',' || *ptr == ';')    /* ap_get_token() bug */
                ptr++;
 
            if (*pair == ' ')
                pair++;
 
            value = pair;
            name = ap_getword(r->pool, &pair, '=');

            if (a->dir_conf->cache_all_cookies) {
                found = 1;

            } else {
                for (i = 0; i < a->dir_conf->cache_cookies->nelts; i++) {
                    if (cookie[i].regexp) {
                        if (ap_regexec(cookie[i].regexp, name, 0, NULL, 0)
                            == 0)
                        {
                            found = 1;
                            break;
                        }

                    } else {
                        if (strcmp(name, cookie[i].name) == 0) {
                            found = 1;
                            break;
                        }
                    }
                }
            }

            if (found) {
                int disabled = 0;

                for (i = 0; i < a->dir_conf->no_cache_cookies->nelts; i++) {

                    if (no_cookie[i].regexp) {
                        if (ap_regexec(no_cookie[i].regexp, name, 0, NULL, 0)
                            == 0)
                        {
                            disabled = 1;
                            break;
                        }

                    } else {
                        if (strcmp(name, no_cookie[i].name) == 0) {
                            disabled = 1;
                            break;
                        }
                    }
                }

                if (!disabled) {
                    const char **new = ap_push_array(cookies);
                    *new = value;
                }
            }
        }

        if (cookies->nelts) {
            if (cookies->nelts > 1) {
                char **cookie = (char **) cookies->elts;
                qsort(cookies->elts, cookies->nelts, sizeof(char *),
                      &accel_compare_cookie);

                for (i = 0; i < cookies->nelts; i++)
                    a->cookie = a->cookie ? ap_pstrcat(r->pool, a->cookie,
                                                       "; ", cookie[i], NULL):
                                            (char *) cookie[i];

            } else {
                a->cookie = *((char **) cookies->elts);
            }
        }

        if (a->cookie)
            a->key = ap_pstrcat(r->pool, a->url, " ", a->cookie, NULL);
    }

    a->md5 = ap_md5(a->r->pool, (unsigned char *) a->key);

    n = m = 0;
    filename[n++] = '/';
    for (i = 0; i < a->server_conf->levels->nelts; i++) {
        for (j = level[i]; j > 0; j--)
            filename[n++] = a->md5[m++];

        filename[n++] = '/';
    }
    while (m < MD5_LEN + 1)
        filename[n++] = a->md5[m++];

    a->cachefile = ap_pstrcat(a->r->pool,
                              a->server_conf->cache_root, filename, NULL);
}

void accel_create_cache_path(accel_rec *a)
{
    char *end = strchr(a->cachefile, '/');

    while (end = strchr(++end, '/')) {
        *end = '\0';
        if (mkdir(a->cachefile, S_IREAD | S_IWRITE | S_IEXEC) < 0
            && errno != EEXIST)
        {
            ap_log_rerror(APLOG_MARK, APLOG_ERR, a->r,
                          "accel: can't creating cache directory %s",
                          a->cachefile);
            a->cache_error = 1;
        }
        *end = '/';
    }
}

int accel_update_cachefile(accel_rec *a)
{
    a->cache_state_line = ap_psprintf(a->r->pool, CACHE_STATE_WRITE_FORMAT,
                                      a->cache_date, a->cache_lm,
                                      a->cache_expires, a->cache_flags,
                                      a->cache_len, a->time);

    lseek(a->cache_fd, (off_t) 0, SEEK_SET);
    if (write(a->cache_fd, a->cache_state_line, CACHE_STATE_LINE_SIZE)
           != CACHE_STATE_LINE_SIZE) {
        ap_log_rerror(APLOG_MARK, APLOG_ERR, a->r,
                      "accel: can't update cache file %s", a->cachefile);
        a->cache_error = 1;
        return FALSE;
    }

    return TRUE;

}

int accel_unlink_cachefile(accel_rec *a)
{
#ifdef MOD_ACCEL_DEBUG
    if (a->cachefile_exists)
        ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, a->r,
                      "accel: unlink cache file %s", a->cachefile);
#endif

    if (a->cachefile_exists && unlink(a->cachefile) == -1) {
        ap_log_rerror(APLOG_MARK, APLOG_ERR, a->r,
                      "accel: error unlinking cache file %s", a->cachefile);
        return -1;
    }

    a->cachefile_exists = 0;
    return 0;
}
