/*
 * Copyright (C) 2001 Igor Sysoev.
 */

#include "mod_accel.h"

/*
   Default backend connector - connects to HTTP backend using URL
   http://hostname[:port][/[path]]

   Returns backend_rec* on success or NULL otherwise
*/

backend_rec *accel_connect_to_http_backend(backend_connect *bc)
{
    int                 sock, i, rc = 0;
    backend_rec        *b;
    struct hostent     *h;
    struct sockaddr_in  addr;

    request_rec *r = bc->r;

    memset(&addr, '\0', sizeof(addr));

    /* first try */
    if (!*bc->backend) {
        if ((b = accel_parse_http_url(r->pool, bc->url)) == NULL) {
            ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r,
                          "accel: invalid backend URL: %s", bc->url);
            return NULL;
        }

        if (bc->addr.s_addr) {
            addr.sin_addr.s_addr = bc->addr.s_addr;

        } else {
            if ((addr.sin_addr.s_addr = ap_inet_addr(b->host)) == INADDR_NONE) {

                if ((h = ap_pgethostbyname(r->pool, b->host)) != NULL
                     && h->h_addr_list[0] != NULL) {

                    addr.sin_addr = *(struct in_addr *)(h->h_addr_list[0]);

                    for (i = 0; h->h_addr_list[i] != NULL; i++) {
                        struct in_addr *new = ap_push_array(bc->backends);
                        *new = *(struct in_addr *)(h->h_addr_list[i]);
                    }

                } else {
                    ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r,
                                  "accel: backend %s not found", b->host);
                    return NULL;
                }
            }
        }

    } else {
        b = bc->b;
    }

    bc->path = b->path;
    bc->port_str = b->port_str;

    addr.sin_family = AF_INET;
    addr.sin_port = htons(b->port);

    while (*bc->backend == 0 || *bc->backend < bc->backends->nelts) {

        if (*bc->backend)
            addr.sin_addr =
                      ( (struct in_addr *) bc->backends->elts )[*bc->backend];

        (*bc->backend)++;

        if ((sock = accel_socket(r, bc->so_rcvbuf)) == -1) {
            bc->backends->nelts = 0;
            return NULL;
        }

#ifdef MOD_ACCEL_DEBUG
        ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, r,
                      "accel: connect to to %s:%d",
                      inet_ntoa(addr.sin_addr), b->port);
#endif

        if ((rc = accel_connect_to_backend(r, sock, &addr, bc->timeout)) >= 0) {
            b->sock = sock;
            b->addr = addr.sin_addr;

#ifdef MOD_ACCEL_DEBUG
            ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, r,
                          "accel: connected to %s:%d",
                          inet_ntoa(b->addr), b->port);
#endif
            return b;

        } else {
            backend_failed *failed = ap_push_array(bc->failed);
            failed->state = (rc == -2) ? HTTP_GATEWAY_TIME_OUT:
                                         HTTP_BAD_GATEWAY;
            failed->addr = addr.sin_addr;

            ap_log_rerror(APLOG_MARK, APLOG_ERR, r,
                          "accel: connect to %s:%d failed while processing %s",
                          inet_ntoa(addr.sin_addr),
                          ntohs(addr.sin_port),
                          bc->url);
        }

        ap_pclosef(r->pool, sock);
    }

    return NULL;
}


static void connect_timeout(int sig)
{
}

int accel_connect_to_backend(request_rec *r,
                             int sock, struct sockaddr_in *addr, int timeout)
{
    int rc;

    if (timeout > 0)
        ap_set_callback_and_alarm(connect_timeout, timeout);
    rc = connect(sock, (struct sockaddr *) addr, sizeof(struct sockaddr_in));
    if (timeout > 0)
        ap_set_callback_and_alarm(NULL, 0);

    if (rc == -1 && errno == EINTR) {
        rc = -2;
        errno = ETIMEDOUT;
    }

    return rc;
}

int accel_socket(request_rec *r, int so_rcvbuf)
{
    int sock = ap_psocket(r->pool, PF_INET, SOCK_STREAM, IPPROTO_TCP);
    if (sock == -1) {
        ap_log_rerror(APLOG_MARK, APLOG_ERR, r, "accel: error creating socket");
        return -1;
    }

    if (so_rcvbuf) {
#ifdef MOD_ACCEL_DEBUG
        ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, r,
                      "SO_RCVBUF %d", so_rcvbuf);
#endif
        if (setsockopt(sock, SOL_SOCKET, SO_RCVBUF,
                   (const char *) &so_rcvbuf, sizeof(int))
                == -1)
            ap_log_rerror(APLOG_MARK, APLOG_ERR, r,
                          "accel: setsockopt(SO_RCVBUF): failed "
                          "to set BkRcvSockBuffSize, using default");
    }

    return sock;
}
