/*
 * Copyright (C) 2001 Igor Sysoev.
 */

#include "mod_accel.h"
#include <time.h>

static time_t now;
static int done;
static int files;
static int deleted;
static double size;
static double freed;

int accel_gc_init(void *data)
{
    accel_gc_conf     *agc = data;
    server_rec        *s = agc->server;
    accel_server_conf *conf = agc->conf;

    /* don't collect garbage during initial test restart */
    if (agc->ppid != 1) {
        sleep(60);
        return -1;
    }

    if (setgid(ap_group_id) == -1) {
        ap_log_error(APLOG_MARK, APLOG_ALERT, s,
                     "accel gc: unable to set group id to %u",
                      (unsigned)ap_group_id);
        return -1;
    }

    if (setuid(ap_user_id) == -1) {
        ap_log_error(APLOG_MARK, APLOG_ALERT, s,
                     "accel gc: unable to set user id to %u",
                     (unsigned)ap_user_id);
        return -1;
    }

    /* wait a minute before collect garbage */
    sleep(60);

    ap_setproctitle("garbage collector");

    for (;;) {
        int        tz, day_time;
        struct tm *tm;

        if (conf->clean_at_time != UNSET) {
            tm = ap_get_gmtoff(&tz);
            day_time = tm->tm_hour * 3600 + tm->tm_min * 60 + tm->tm_sec;
            sleep(conf->clean_at_time > day_time ?
                                  conf->clean_at_time - day_time:
                                  24 * 3600 - day_time + conf->clean_at_time);
        }

        done = 0;
        files = deleted = 0;
        size = freed = 0;
        now = time(NULL);

        ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_NOTICE, s,
                     "accel gc: collect garbage");

        accel_collect_garbage(s, conf->cache_root, conf->levels, 0,
                              conf->clean_la);

        ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_NOTICE, s,
                     "accel gc: files: %d total, %d deleted", files, deleted);
        ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_NOTICE, s,
                     "accel gc: cache: %.3fM total, %.3fM freed",
                      size / 1024 / 1024 , freed / 1024 / 1024);

        if (conf->clean_at_time == UNSET)
            sleep(conf->clean_time);
    }
}


void accel_collect_garbage(server_rec *s, char *dirname,
                           array_header *levels, int nlevel, int max_la)
{
    int           *level, fd, date, lm, expires, flags, len, la;
    DIR           *dir;
    struct dirent *de;
    struct stat    sb;
    char           state[CACHE_STATE_LINE_SIZE + 1];

    char *filename = NULL;
    int dirnamelen = strlen(dirname);

    if (!(dir = opendir(dirname))) {
        ap_log_error(APLOG_MARK, APLOG_ERR, s,
                     "accel gc: can't open directory %s", dirname);
        return;
    }

    while ((de = readdir(dir)) != NULL) {

#ifdef __FreeBSD__
        int namelen = de->d_namlen;
#else
        int namelen = strlen(de->d_name);
#endif

        if (filename)
            free(filename);

        filename = malloc(dirnamelen + namelen + 2);
        *filename = '\0';
        strcpy(filename, dirname);
        filename[dirnamelen] = '/';
        filename[dirnamelen + 1] = '\0';
        strcat(filename, de->d_name);

        done++;
        if (done > 1000) {
            done = 0;
            sleep(1);
            now = time(NULL);
        }

        if (stat(filename, &sb) == -1) {
            if (errno != ENOENT)
                ap_log_error(APLOG_MARK, APLOG_ERR, s,
                             "accel gc: can't stat file %s", filename);
            continue;
        }

        if (S_ISDIR(sb.st_mode)) {

#ifdef MOD_ACCEL_DEBUG0
            ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, s,
                         "accel gc: enter directory %s",
                         filename);
#endif

            if (strcmp(de->d_name, ".") == 0 || strcmp(de->d_name, "..") == 0)
                continue;

            size += sb.st_size;
            level = (int *) levels->elts;

#ifdef MOD_ACCEL_DEBUG0
            if (nlevel == -1)
                ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, s,
                             "accel gc: level is -1");
            else
                ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, s,
                             "accel gc: level %d:%d:%d:%d",
                             nlevel, namelen,
                             level[nlevel], levels->nelts);
#endif

            if (nlevel == -1
                || nlevel == levels->nelts
                || (nlevel < levels->nelts && namelen != level[nlevel]))
            {
#ifdef MOD_ACCEL_DEBUG
                ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, s,
                             "accel gc: enter garbage directory %s",
                             filename);
#endif

                accel_collect_garbage(s, filename, levels, -1, max_la);

                ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_NOTICE, s,
                             "accel gc: unlink garbage directory %s",
                             filename);

                freed += sb.st_size;

                if (rmdir(filename) == -1)
                    ap_log_error(APLOG_MARK, APLOG_ERR, s,
                              "accel gc: can't unlinking garbage directory %s",
                              filename);

                continue;
            }

            accel_collect_garbage(s, filename, levels, nlevel + 1, max_la);

        } else {  /* files */

            int error = 0;

            files++;
            size += sb.st_size;

            if (nlevel == 0
                && strncmp(de->d_name, "tmp", 3) == 0) {

                if (now - sb.st_atime > TOO_LONG_TIME
                    && now - sb.st_mtime > TOO_LONG_TIME)
                {

                    ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_NOTICE, s,
                                 "accel gc: unlink temp file %s",
                                 filename);

                    deleted++;
                    freed += sb.st_size;

                    if (unlink(filename) == -1)
                        ap_log_error(APLOG_MARK, APLOG_ERR, s,
                                     "accel gc: can't unlinking temp file %s",
                                     filename);
                }

                continue;
            }

            if (nlevel != levels->nelts) {
#ifdef MOD_ACCEL_DEBUG
                ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, s,
                             "accel gc: unlink garbage file %s",
                             filename);
#endif

                deleted++;
                freed += sb.st_size;

                if (unlink(filename) == -1)
                    ap_log_error(APLOG_MARK, APLOG_ERR, s,
                                 "accel gc: can't unlinking garbage file %s",
                                 filename);

                continue;
            }

#ifdef MOD_ACCEL_DEBUG
            ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, s,
                         "accel gc: check file %s", filename);
#endif
            state[CACHE_STATE_LINE_SIZE] = '\0';

            if (((fd = open(filename, O_RDONLY | O_BINARY, 0640)) < 0
                  || read(fd, state, CACHE_STATE_LINE_SIZE)
                                  != CACHE_STATE_LINE_SIZE))
            {
                ap_log_error(APLOG_MARK, APLOG_ERR, s,
                             "accel gc: can't read from file %s", filename);
                error = 1;

            } else if (!ap_checkmask(state, cache_state_line_mask)) {
                ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, s,
                             "accel gc: can't read from file %s", filename);
                error = 1;
            }

            close(fd);

            if (error) {
                deleted++;
                freed += sb.st_size;

                if (unlink(filename) == -1)
                    ap_log_error(APLOG_MARK, APLOG_ERR, s,
                                 "accel gc: can't unlinking stale file %s",
                                 filename);
                continue;
            }

            sscanf(state, CACHE_STATE_GC_FORMAT,
                   &date, &lm, &expires, &flags, &len, &la);

            if (now - expires > TOO_LONG_TIME || now - la > max_la) {
#ifdef MOD_ACCEL_DEBUG
                ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, s,
                             "accel gc: unlink stale file %s",
                             filename);
#endif

                deleted++;
                freed += sb.st_size;

                if (unlink(filename) == -1)
                    ap_log_error(APLOG_MARK, APLOG_ERR, s,
                                 "accel gc: can't unlinking stale file %s",
                                 filename);
            }
        }
    }

    if (filename)
        free(filename);

    closedir(dir);
}
