/*
 * Copyright (C) 2001 Igor Sysoev.
 */

#include "mod_accel.h"

int accel_read_request_body(accel_rec *a)
{
    int pos;
    int len = 1;

    request_rec *r = a->r;
    accel_dir_conf *dcf = a->dir_conf;

    if (r->remaining <= dcf->request_buff_size) {

        for (pos = 0; len > 0; pos += len) {
            ap_soft_timeout("accel: read request data from client", r);
            len = ap_get_client_block(r, a->request_buf + pos,
                                      dcf->request_buff_size - pos);
            ap_kill_timeout(r);

            if (len == -1) {
                if (!r->connection->aborted)
                    ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_INFO, r,
                                  "accel: client stopped request data sending");
                return HTTP_BAD_REQUEST;
            }
        }

    } else {
        char *rc;

        if ((rc = accel_create_tempfile(a, &a->request_file, &a->request_fd))
                != NULL) {
            ap_log_rerror(APLOG_MARK, APLOG_ERR, r, rc);
            return HTTP_INTERNAL_SERVER_ERROR;
        }

        unlink(a->request_file);

        for (pos = 0; len > 0; ) {
            ap_soft_timeout("accel: read request data from client", r);
            len = ap_get_client_block(r, a->request_buf + pos,
                                      dcf->request_buff_size - pos);
            ap_kill_timeout(r);

            if (len == -1) {
                if (!r->connection->aborted)
                    ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_INFO, r,
                                  "accel: client stopped request data sending");
                return HTTP_BAD_REQUEST;
            }

            pos += len;

            if (pos == dcf->request_buff_size || len == 0) {
                if (write(a->request_fd, a->request_buf, pos) != pos) {
                    ap_log_rerror(APLOG_MARK, APLOG_ERR, r,
                                  "accel: can't write to tempfile %s",
                                  a->request_file);
                    return HTTP_INTERNAL_SERVER_ERROR;
                }

                a->request_seek += pos;
                pos = 0;
            }
        }

    }

    return OK;
}

void accel_request_body_done(accel_rec *a)
{
    if (a->request_fd >= 0)
        ap_pclosef(a->r->pool, a->request_fd);

    accel_freea(a->request_buf);
}


/*
   Read HTTP response header from backend's BUFF* or cache's FILE*.
   Set a->(backend|cache)_header and return TRUE on success or FALSE otherwise.

   Read HTTP/1.0+ header only.
*/

int accel_read_http_header(accel_rec *a, int source)
{
    int          len;
    char         buf[HUGE_STRING_LEN];
    header_rec  *h;

    request_rec *r = a->r;

    if ((len = accel_gets(buf, HUGE_STRING_LEN, a, source)) < 0) {
        accel_gets_error(a, source, 1, len, errno);

        if (source == BACKEND)
            ap_bclose(a->backend_fb);
        else
            ap_pclosef(r->pool, a->cache_fd);

        return FALSE;
    }

    if (!ap_checkmask(buf, "HTTP/#.# ###*")) {

        if (source == BACKEND) {
            ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r,
                          "accel: not HTTP/1.0+ response from "
                          "backend %s while processing %s",
                          a->backend_ip, a->url);
        } else { 
            ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r,
                          "accel: not HTTP/1.0+ response from cache %s",
                          a->cachefile);
        }

        return FALSE;
    }

    h = ap_pcalloc(r->pool, sizeof(header_rec));
    h->line = ap_pstrdup(r->pool, buf);
    h->status_line = ap_pstrdup(r->pool, &buf[9]);
    buf[12] = '\0';
    h->status = atoi(&buf[9]);

    if (!(h->headers_out = accel_read_http_headers(a, source))) {
        if (source == BACKEND) {
            ap_bclose(a->backend_fb);
            ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r,
                          "accel: not valid HTTP/1.0+ header from "
                          "backend %s while processing %s",
                          a->backend_ip, a->url);
        } else {
            ap_pclosef(r->pool, a->cache_fd);
            ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r,
                          "accel: not valid HTTP/1.0+ header from cache %s",
                          a->cachefile);
        }

        return FALSE;
    }

    if (source == BACKEND) {
        a->backend_header = h;
    } else {
        a->cache_header_size = a->temp_last - a->temp_buf;
        a->cache_header = h;
    }

    return TRUE;
}

/*
    Set r->status, r->status_line, r->content_type, r->content_encoding
        and r->headers_out from a->(backend|cache)_header.

    Remove "X-Accel-*" private headers.
    Remove "X-Pad" header - header size is changed and Apache core
                            set it if it's needed.
    Remove "Connection" header - it should be set by Apache core.
    Remove "Server" header if AccelPassServer is off - Apache 1.3.24 allows
                            to set own "Server" header.
    Set correct "Expires" header if response is cached.
    Rewrite "Cache-Control: max-age=#" header.

    Remove backend's "Set-Cookie" header if request is cachable
                                            and AccelPassCookie is off
           and set local "Set-Cookie" header if exists.

    Rewrite "Location" and "Refresh" headers.
*/

void accel_set_http_header(accel_rec *a, int status)
{
    int          expires, port, rewrite = 0;
    const char  *cookie, *ptr, *range = NULL, *len = NULL;

    request_rec *r = a->r;
    header_rec  *h = a->header == BACKEND ? a->backend_header:
                                            a->cache_header;
    switch (status) {

    case HTTP_NOT_MODIFIED:
        r->status = HTTP_NOT_MODIFIED;
        r->status_line = "304 Not Modified";
        break;

    case HTTP_PRECONDITION_FAILED:
        r->status = HTTP_PRECONDITION_FAILED;
        r->status_line = "412 Precondition Failed";
        break;

    case HTTP_PARTIAL_CONTENT:
        r->status = HTTP_PARTIAL_CONTENT;
        r->status_line = "206 Partial Content";
        range = ap_table_get(r->headers_out, "Content-Range");
        len = ap_table_get(r->headers_out, "Content-Length");
        break;

    case HTTP_RANGE_NOT_SATISFIABLE:
        r->status = HTTP_RANGE_NOT_SATISFIABLE;
        r->status_line = "416 Requested Range Not Satisfiable";
        range = ap_table_get(r->headers_out, "Content-Range");
        break;

    default:
        r->status = h->status;
        r->status_line = h->status_line;
    }

    r->content_type = ap_pstrdup(r->pool,
                            ap_table_get(h->headers_out, "Content-Type"));
    r->content_encoding = ap_pstrdup(r->pool,
                            ap_table_get(h->headers_out, "Content-Encoding"));

    /* BUG: get single cookie line only */
    cookie = ap_table_get(r->headers_out, "Set-Cookie");

    r->headers_out = ap_copy_table(r->pool, h->headers_out);

    /* X-Accel-* headers */
    if (!a->dir_conf->pass_x_accel)
        ap_table_unset(r->headers_out, "X-Accel-Expires");

    ap_table_unset(r->headers_out, "X-Pad");
    ap_table_unset(r->headers_out, "Connection");

    if (!a->dir_conf->pass_server)
#if APACHE_RELEASE < 10325000
        ap_table_unset(r->headers_out, "Server");
#else
        ap_table_set(r->headers_out, "Server", ap_get_server_version());
#endif

    /* Expires */
    expires = ap_parseHTTPdate(ap_table_get(r->headers_out, "Expires"));
    if (a->header == CACHE && expires)
    {
        ap_table_set(r->headers_out, "Expires",
                     ap_gm_timestr_822(r->pool, a->cache_expires));
    }

    /* Cache-Control: max-age=# */
    if (ptr = ap_table_get(r->headers_out, "Cache-Control")) {
        const char *token;
        char *line = "";
        char *sep = "";
        int max_age = -1;

        while (*ptr && (token = ap_get_token(r->pool, &ptr, 0))) {
            if (*ptr == ',' || *ptr == ';')         /* ap_get_token() bug */
                ptr++;

            /* BUG: look "max-age=x", but may be "max-age = " */
            if (strncmp(token, "max-age=", 8) == 0) {
                if (a->header == CACHE) {
                    if ((max_age = a->cache_expires - r->request_time) < 0)
                        max_age = 0;
                } else if (expires) {
                    if ((max_age = expires - r->request_time) < 0)
                        max_age = 0;
                }

                token = ap_psprintf(r->pool, "max-age=%d", max_age);
            }

            line = ap_pstrcat(r->pool, line, sep, token, NULL);
            sep = "; ";
        }

        if (max_age != -1)
            ap_table_set(r->headers_out, "Cache-Control", line);
    }

    if (range)
        ap_table_setn(r->headers_out, "Content-Range", range);

    if (len)
        ap_table_setn(r->headers_out, "Content-Length", len);

    /* Set-Cookie */
    if (is_cachable(a) && !a->dir_conf->pass_cookie)
        ap_table_unset(r->headers_out, "Set-Cookie");

    if (cookie)
        ap_table_addn(r->headers_out, "Set-Cookie", cookie);

    if (a->pass->preserve_host)
        rewrite = (port = ap_get_server_port(r)) != a->pass->port;
    else
        rewrite = 2;

    if (rewrite) {
        char *url, *refresh, *new;

        /* Location */
        if (url = (char *)ap_table_get(r->headers_out, "Location")) {
            new = (rewrite == 2) ? accel_rewrite_url(a, url):
                                   accel_rewrite_port(a, url, port);
            if (new)
                ap_table_setn(r->headers_out, "Location", new);
        }

        /* Refresh */
        if ((refresh = (char *)ap_table_get(r->headers_out, "Refresh"))
            && (url = strstr(refresh, "url=")))
        {
            new = (rewrite == 2) ? accel_rewrite_url(a, url + 4):
                                   accel_rewrite_port(a, url + 4, port);
            if (new)
                ap_table_setn(r->headers_out, "Refresh",
                              ap_pstrcat(r->pool,
                                         ap_pstrndup(r->pool,
                                                   refresh, url + 4 - refresh ),
                                         new, NULL));
        }
    }
}

/*
   Read response headers until empty CRLF string has received.
   Return string without '\n'.

   buf is very big for one header string (HUGE_STRING_LEN, by default is 8K)
   even it's multi-line one so we should get '\n' in the end of the string
   otherwise it's not valid header string.
*/

table *accel_read_http_headers(accel_rec *a, int source)
{
    int    len, rc;
    char   buf[HUGE_STRING_LEN];
    char   next, *value, *end;

    table *headers = ap_make_table(a->r->pool, 20);

    char  *pos = buf;
    int    n = HUGE_STRING_LEN;

    for (;;) {
        len = accel_gets(pos, n, a, source);

        /* Error, EOF or no CRLF in end of line */
        if (len < 0) {
            accel_gets_error(a, source, 0, len, errno);
            return NULL;
        }

        /* CRLF - end of response headers */
        if (len == 0)
            return headers;

        /* Check multi-line string form backend.
           Cache files have not multi-linestrings */
        if (source == BACKEND) {
            if ((rc = accel_lookc(&next, a)) != 1) {
                accel_gets_error(a, source, 0, rc, errno);
                return NULL;
            }

            if (next == ' ' || next == '\t') {
                pos += len;
                n -= len;
                continue;
            }
        }

#ifdef MOD_ACCEL_DEBUG
        ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, a->r,
                      "backend header: '%s'", buf);
#endif

        if (!(value = strchr(buf, ':'))) {      /* not "Key: Value"          */
            if (strncmp(buf, "HTTP/", 5) == 0) { /* IIS bug  */
                pos = buf;
                n = HUGE_STRING_LEN;
                continue;
            }

            ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, a->r,
                          "accel: invalid header '%s' from "
                          "backend %s while processing %s",
                          buf, a->backend_ip, a->url);
            return NULL;
        }

        *(value++) = '\0';                      /* end of key                */

        while (*value == ' ' || *value == '\t') /* skip leading spaces       */
            ++value;

        for (end = pos + len - 1;               /* skip trailing spaces      */
                end > value
                    && (*end == ' ' || *end == '\t');
                --end)
            *end = '\0';

        ap_table_add(headers, buf, value);

        pos = buf;
        n = HUGE_STRING_LEN;
    }
}

/*
    Parse HTTP URL and return host, port, port_str and path in backend_rec.
*/

backend_rec *accel_parse_http_url(pool *p, char *url)
{
    char        *host, *port, *path;

    backend_rec *b = ap_pcalloc(p, sizeof(backend_rec));

    if (!url || strncasecmp(url, "http://", 7) != 0)
        return NULL;

    host = url + 7;
    if (path = strchr(host, '/')) {
        b->host = ap_pstrndup(p, host, path - host);
        b->path = ap_pstrdup(p, path);

    } else {
        b->host = ap_pstrdup(p, host);
        b->path = "/";
    }

    if (port = strchr(b->host, ':')) {
        *(port++) = '\0';
        if (!ap_isdigit(*port))
            return NULL;

        if ((b->port = atoi(port)) != DEFAULT_HTTP_PORT) {
            b->port_str = ap_pstrcat(p, ":", port, NULL);
            return b;
        }

    }

    b->port_str = "";
    b->port = DEFAULT_HTTP_PORT;

    return b;
}
