/*
 * Copyright (C) 2001 Igor Sysoev.
 */

#ifndef EAPI
#error "mod_accel requires Extended API (EAPI)"
#endif

#if !defined(EAPI_MM ) && defined(__GNUC__)
#warning "compiled without MM library -- busy locks would not be supported"
#endif

#include "httpd.h"
#include "http_core.h"
#define CORE_PRIVATE
#include "http_config.h"
#undef CORE_PRIVATE
#include "http_conf_globals.h"
#include "http_protocol.h"
#include "http_request.h"
#include "http_log.h"
#include "http_main.h"
#include "util_date.h"
#include "util_md5.h"
#include "scoreboard.h"
#include "fnmatch.h"

#include "ap_ext_accel.h"
#include "ap_other_child.h"

#ifndef NO_WRITEV
#include <sys/types.h>
#include <sys/uio.h>
#endif

#include "ap_accel.h"

#define MOD_ACCEL_VER           "mod_accel/1.0.25"

#define UNSET                   -1

/* accel_gets error codes */
#define ACCEL_GETS_TOO_LONG     -3
#define ACCEL_GETS_ERROR        -2
#define ACCEL_GETS_EOF          -1

/* backend cache-control */
#define ACCEL_NO_CACHE          -2

/* lock state flags */
#define ACCEL_UPDATING          1
#define ACCEL_MAX_CONN          2
#define ACCEL_MAX_WAIT          4
#define ACCEL_NEW_HIT           8

#define TEMP_SIZE               8192
#define REQUEST_BUFF_SIZE       8192
#define BK_RCV_BUFF_SIZE        16384
#define SEND_SIZE               8192
#define CACHE_SEND_SIZE         8192

#define MD5_LEN                 32

#define BACKEND                 0
#define CACHE                   1

#define ACCEL_NO_STUB           0
#define ACCEL_STUB              1

typedef enum {
    ACCEL_MGR_NO_ACTION = 0,
    ACCEL_MGR_REFRESH,
    ACCEL_MGR_CREATE,
    ACCEL_MGR_REMOVE
} accel_mgr_e;

typedef enum {
    ACCEL_MGR_SUCCESS = 0,
    ACCEL_MGR_INVALID,
    ACCEL_MGR_NO_ACCELERATED,
    ACCEL_MGR_NOT_FOUND,
    ACCEL_MGR_NO_CACHABLE,
    ACCEL_MGR_CACHE_ERROR,
    ACCEL_MGR_BACKEND_ERROR
} accel_mgr_status_e;

typedef enum {
    ACCEL_SUM_NONE = 0,
    ACCEL_SUM_HOST,
    ACCEL_SUM_PORT,
    ACCEL_SUM_LOCATION,
    ACCEL_SUM_TAG
} accel_sum_e;

typedef enum {
    ACCEL_BL_ERR = 0,
    ACCEL_BL_NONE,
    ACCEL_BL_OLD,
    ACCEL_BL_OK
} accel_bl_e;

#define TOO_LONG_TIME           3600

typedef enum {
    ACCEL_PASS = 1,
    ACCEL_NTNC,
    ACCEL_AUTH,
    ACCEL_RMVD,
    ACCEL_INVL,
    ACCEL_PGNC,
    ACCEL_MISS,
    ACCEL_AGED,
    ACCEL_EXPR,
    ACCEL_RVUS,
    ACCEL_HIT
} accel_state_e;

typedef enum {
    ACCEL_NNC = 1,
    ACCEL_XAE,
    ACCEL_CTL,
    ACCEL_EXP,
    ACCEL_MVD,
    ACCEL_LMF,
    ACCEL_ADE
} backend_cache_e;

#define is_cachable(a)           (a->accel_state >= ACCEL_INVL)
#define is_max_connected(a)      (a->max_connected || a->max_waiting)

#define CACHE_STATE_LINE_MASK    \
              "&&&&&&&& &&&&&&&& &&&&&&&& &&&&&&&& &&&&&&&& &&&&&&&&"
#define CACHE_LEN_POS            ((8 + 1) * 4)
#define CACHE_STATE_READ_FORMAT  "%8x %8x %8x %8x %8x"
#define CACHE_STATE_GC_FORMAT    "%8x %8x %8x %8x %8x %8x"
#define CACHE_STATE_WRITE_FORMAT "%08x %08x %08x %08x %08x %08x"

#define CACHE_STATE_LINE_SIZE    (sizeof(CACHE_STATE_LINE_MASK) - 1)


#ifdef NO_ALLOCA
#define accel_alloca(size)       accel_alloc(a->allocated, size)
#define accel_freea(p)           accel_free(a->allocated, p);
#else
#define accel_alloca(size)       alloca(size)
#define accel_freea(p)
#endif

#define init(field, default)     if (field == UNSET) field = default

#define merge(merged, parent, child, field)                              \
              merged->field = child->field == UNSET ?                    \
                                parent->field : child->field

#define accel_make_url(r, uri)                                           \
              ap_pstrcat(r->pool, uri,                                   \
                            /*  r->path_info ? r->path_info : "", */     \
                                r->args ? "?" : "",                      \
                                r->args ? r->args : "",                  \
                                NULL)

typedef struct {
    int          max_conn;
    int          max_wait;
    int          sum_part;
    int          sum;
    int          loc_len;
    int          url_len;
    char        *loc;
    char        *url;
    int          port;
    unsigned     preserve_host:1;
    unsigned     cache_host:1;
    unsigned     the_same_host:1;
    struct in_addr addr;
} accel_pass;

typedef struct {
    int          len;
    char        *loc;
    regex_t     *regexp;
} accel_no_pass;

typedef struct {
    int          loc_len;
    int          url_len;
    char        *loc;
    char        *url;
} accel_reverse;

typedef struct {
    char        *name;
    regex_t     *regexp;
} accel_cookie;


/* Busy lock state and slot */

typedef enum {
    ACCEL_FREE = 0,                      /* free slot                        */
    ACCEL_BUSY_LOCK,                     /* request is processing with b/l   */
    ACCEL_BUSY_PASS,                     /* request is processing w/o b/l    */
    ACCEL_WAIT                           /* child is waiting in b/l          */
} accel_slot_e;

typedef struct {
    accel_slot_e state;
    time_t       time;
    char         md5[MD5_LEN + 1];
    int          sum;
} accel_slot;



typedef struct {
    array_header *pass;                  /* AccelPass                        */
    array_header *no_pass;               /* AccelNoPass                      */
    array_header *reverse;               /* AccelReverse                     */
    char         *cache_root;            /* AccelCacheRoot                   */
    array_header *levels;                /* -- cache levels                  */
    int           auto_create_path;      /* -- noauto                        */

    int           clean_at_time;         /* AccelClean                       */
    int           clean_time;            /* --                               */
    int           clean_la;              /* AccelCleanLastAccessed           */
} accel_server_conf;

typedef struct {
    int           unlink_no_cached;      /* AccelUnlinkNoCached              */
    int           no_cache;              /* AccelNoCache                     */
    int           ignore_no_cache;       /* AccelIgnoreNoCache               */
    int           ignore_auth;           /* AccelIgnoreAuth                  */
    int           revalidate_user;       /* AccelRevalidateUser              */
    int           pass_cookie;           /* AccelPassCookie                  */
    int           cache_set_cookie;      /* AccelCacheSetCookie              */

    int           cache_cookie;          /* AccelCacheCookie                 */
    int           cache_all_cookies;     /* --                               */
    array_header *cache_cookies;         /* --                               */
    array_header *no_cache_cookies;      /* --                               */

    int           request_buff_size;     /* AccelRequestBuffSize             */
    int           bk_rcv_buff_size;      /* AccelBkRcvBuffSize               */
    int           bk_rcv_sock_buff_size; /* AccelBkRcvSockBuffSize           */
    int           timeout;               /* AccelTimeout                     */
    int           connect_timeout;       /* -- connect timeout               */
    int           retry_5XX;             /* AccelRetry5XX                    */
    int           wait_before_body_read; /* AccelWaitBeforeBodyRead          */
    int           send_size;             /* AccelSendSize                    */
    int           cache_send_size;       /* AccelCacheSendSize               */

    int           cache_lm_factor;       /* AccelLastModifiedFactor          */
    int           default_expire;        /* AccelDefaultExpire               */
    int           ignore_expires;        /* AccelIgnoreExpires               */
    char         *invalidate;            /* AccelInvalidate                  */
    int           invl_len;

    int           x_fwd_for;             /* AccelAddXForwardedFor            */
    int           via;                   /* AccelAddVia                      */
    int           x_real_ip;             /* AccelSetXRealIP                  */
    int           x_host;                /* AccelSetXHost                    */
    int           x_url;                 /* AccelSetXURL                     */
    int           x_uri;                 /* AccelSetXURI                     */
    int           pass_server;           /* AccelPassServer                  */
    int           pass_x_accel;          /* AccelPassXAccel                  */

    int           max_stale;             /* AccelMaxStale                    */
    int           busy_lock;             /* AccelBusyLock                    */
    int           expired_busy_lock;     /* -- expired busy lock             */
    int           max_busy_lock;         /* -- max busy lock                 */
} accel_dir_conf;

/* accel_rec */

typedef struct accel_buff accel_buff_rec;

typedef struct {
    accel_server_conf   *server_conf;
    accel_dir_conf      *dir_conf;
    request_rec         *r;
    accel_buff_rec      *ab;

    array_header *allocated;
    char        *timeout_name;

    accel_pass  *pass;
    char        *url;                    /* backend url                      */
    char        *rest;                   /* rest part in the url             */

    char        *request_buf;
    char        *request_file;
    int          request_fd;
    off_t        request_seek;

    char        *temp_buf;
    char        *temp_last;
    char        *temp_end;
    int          temp_size;

    char        *tempfile;               /* temp filename of received request */
    int          temp_fd;                /* temp file's fd                   */
    off_t        temp_seek;              /* last seek position in temp file,
                                            used to avoid needless lseek()s  */

    time_t       time;                   /* cached time                      */

    char        *cookie;
    char        *key;                    /* URL's key for md5                */
    char        *md5;                    /* key's md5                        */
    char        *cachefile;              /* cache filename                   */
    int          cache_fd;               /* cache's fd                       */
    char        *cache_buf;
    off_t        cache_size;
    off_t        cache_header_size;
    off_t        cache_seek;

    char        *cache_state_line;
    time_t       cache_date;
    time_t       cache_lm;
    time_t       cache_expires;
    int          cache_flags;
    int          cache_len;

    array_header *backends;

    backend_rec *backend;
    char        *backend_ip;
    int          backend_fd;             /* backend's fd                     */
    BUFF        *backend_fb;             /* backend's BUFF*                  */
    BUFF        *client_fb;              /* client's BUFF*                   */

    const char  *method;
    char        *buf;                    /* buffer for received request      */
    int          bk_rcv_buff_size;       /* AccelBkRcvBuffSize, size of *buf */
    int          send_threshold;
    int          header_end;
    int          inptr;                  /* input pointer in *buf            */
    int          outptr;                 /* output pointer in *buf           */
    int          incnt;                  /* bytes received from backend      */
    int          outcnt;                 /* bytes sent to client             */

    /* log values */
    accel_state_e accel_state;
    time_t       expired_time;
    int          lock_state;
    time_t       lock_time;
    int          backend_state;
    int          backend_first_read;
    int          backend_second_read;
    int          backend_read;
    int          backend_reads;
    time_t       backend_time;
    backend_cache_e backend_cache;
    int          backend_cache_time;

    accel_cnt   *cnt;

    header_rec  *backend_header;
    header_rec  *cache_header;

    int          expired;
    int          cachefile_existed;

    unsigned     external:1;            /* 'proxy:http://backend/' */

    unsigned     cachemgr:1;

    unsigned     note_nocache:1;

    unsigned     max_connected_checked:1;
    unsigned     max_connected:1;
    unsigned     max_waiting:1;

    unsigned     header:1;
    unsigned     body:1;

    unsigned     cache_error:1;
    unsigned     cachefile_exists:1;
    unsigned     cachefile_path_exists:1;
    unsigned     revalidate:1;
    unsigned     invalidate:1;
    unsigned     busy_lock:1;            /* busy lock set */

    unsigned     cache_body:1;           /* read whole body to cache it even
                                            request need only header */
    unsigned     client_writable:1;

    unsigned     backend_nonblocked:1;
    unsigned     backend_aborted:1;
    unsigned     backend_timedout:1;
    unsigned     backend_closed:1;
} accel_rec;

/* accel_buff_rec */

struct accel_buff {
    accel_rec   *a;
    table       *log_notes;
};

/* Garbage collector's data */

typedef struct {
    server_rec  *server;
    pool        *pool;
    int          ppid;
    accel_server_conf *conf;
} accel_gc_conf;


/* Forward declaration */

extern module accel_module;

accel_rec *accel_init_ctx(request_rec *r);
void accel_init_dir_config(accel_rec *a, accel_dir_conf *conf);

#ifdef MOD_ACCEL_CACHEMGR
/* accel_cachemgr.c */
int accel_cachemgr(request_rec *r);
#endif

/* accel_backend.c */
int  accel_read_backend_header(accel_rec *a, int *backend);
void accel_clean_backend_state(accel_rec *a);
int  accel_send_request_to_backend(accel_rec *a);
void accel_send_backend_range(accel_rec *a);
void accel_send_backend_body(accel_rec *a);
int  accel_write_hook(BUFF *fb, char *buf, int len);
#ifndef NO_WRITEV
int  accel_writev_hook(BUFF *fb, const struct iovec *iov, int iovcnt);
#endif
void accel_read_and_check_writable(accel_rec *a, BUFF *fb, int hook);
int  accel_backend_read(accel_rec *a);
void accel_close_backend(accel_rec *a);
int  accel_read_from_buffer(accel_rec *a, char *buf, int pos, int len);


/* accel_cache.c */
int  accel_check_cachable(accel_rec *a);
int  accel_check_cache(accel_rec *a);
int  accel_open_cachefile_with_busy_lock(accel_rec *a, int expired);
int  accel_read_cachefile_header(accel_rec *a);
void accel_send_cache_range(accel_rec *a);
void accel_send_cache_body(accel_rec *a);
int  accel_check_ims(accel_rec *a, int last_modified, header_rec *h);
void accel_update_cache(accel_rec *a);
void accel_response_expired(accel_rec *a);
int  accel_compare_cookie(const void *one, const void *two);
void accel_make_cachefile_name(accel_rec *a, int cache_cookie);
void accel_create_cache_path(accel_rec *a);
int  accel_update_cachefile(accel_rec *a);
int  accel_unlink_cachefile(accel_rec *a);


/* accel_lock.c */
int  accel_check_busy_lock(accel_rec *a, int busy_lock, int wait);
void accel_clear_busy_lock(accel_rec *a);
int  accel_max_connected(accel_rec *a);
void accel_set_slot_num(void);


/* accel_gc.c */
int  accel_gc_init(void *data);
void accel_collect_garbage(server_rec *s, char *dirname,
                           array_header *levels, int nlevel, int max_la);


/* accel_connect.c */
backend_rec *accel_connect_to_http_backend(backend_connect *bc);
int          accel_connect_to_backend(request_rec *r, int sock,
                                      struct sockaddr_in *addr, int timeout);
backend_rec *accel_parse_http_url(pool *p, char *url);


/* accel_http.c */
int    accel_read_request_body(accel_rec *a);
void   accel_request_body_done(accel_rec *a);
int    accel_read_http_header(accel_rec *a, int source);
void   accel_set_http_header(accel_rec *a, int status);
table *accel_read_http_headers(accel_rec *a, int source);


/* accel_util.c */
char  *accel_create_tempfile(accel_rec *a, char **tempfile, int *temp_fd);
int    accel_open_tempfile(accel_rec *a, int stub);
int    accel_tempfile_write(accel_rec *a, char *buf, int size, off_t seek);

void   accel_set_timeout(char *timeout_name, accel_rec *a);
void   accel_kill_timeout(accel_rec *a);

void   accel_set_notes(accel_rec *a);
const char *accel_log(request_rec *r, char *a);

int    accel_max_age(pool *p, const char *string);

int    accel_gets(char *buf, int size, accel_rec *a, int source);
int    accel_lookc(char *next, accel_rec *a);
void   accel_gets_error(accel_rec *a, int source, int status,
                        int error, int save_errno);
void   accel_block_error(request_rec *r, int socket, int block);


void  *accel_alloc(array_header *allocated, int size);
void   accel_free(array_header *allocated, void *ptr);
void   accel_cleanup_allocated(array_header *allocated);

void  *accel_set_time_string(pool *p, const char *name,
                             const char *string, int *value);
int    accel_sum(accel_sum_e part, char *url, const char *tag);
char  *accel_rewrite_url(accel_rec *a, char *url);
char  *accel_rewrite_port(accel_rec *a, char *url, int port);


#ifdef EAPI_MM
extern int         accel_slot_num;
extern pool       *shared_pool;
extern accel_slot *accel_state;
#endif

extern char       *cache_state_line_mask;
extern int         accel_content_tail;
