/*
 * Copyright (C) 2002 Igor Sysoev.
 */

/*
MODULE-DEFINITION-START
Name: freeze_module
ConfigStart
ConfigEnd
MODULE-DEFINITION-END
*/

#ifndef EAPI
#error "mod_freeze requires Extended API (EAPI)"
#endif

#include "httpd.h"
#define CORE_PRIVATE
#include "http_config.h"
#undef CORE_PRIVATE
#include "http_log.h"

#include "../accel/ap_accel.h"

#define UNSET   -1
#define MAX_LEN 11      /* "javascript:" */

#define SPACE           (char *)-1
#define OPEN_TAG        (char *)-2
#define CLOSE_TAG       (char *)-3
#define APOST           (char *)-4
#define QUOTE           (char *)-5
#define SLASH           (char *)-6
#define EQUAL           (char *)-7

#define SPACE_DELIM     " \n\r\t"
#define SPACE_CLOSE     "> \n\r\t"
#define SPACE_EQ        "= \n\r\t"


typedef struct {
    int    enable;
    char  *start;
    int    start_len;
} freeze_conf;

typedef struct state_s {
    int                   len;
    char                 *name;
    char                 *delim;
    struct state_s       *s;
    void                (*f)();
} state_r;

module freeze_module;

static void parse();
static void freeze_tag();
static void freeze_param();
static void freeze_scheme();
static void freeze_content_tail_hook(server_rec *s, int *tail_len);
static int  freeze_range_hook(request_rec *r, header_rec *h);
static void freeze_content_hook(request_rec *r, accel_cnt *cnt);

static state_r tag[];
static state_r tag_start[];
static state_r param[];
static state_r active_param[];
static state_r value[];

static state_r tag_begin[] = {
    { 5, NULL, NULL, tag_start, parse },
    { 0, NULL, NULL, tag_begin, parse }
};

static state_r tag_start[] = {
    { 0, OPEN_TAG, NULL, tag, parse },
    { 0, NULL, NULL, tag_start, parse }
};

static state_r tag_end[] = {
    { 0, CLOSE_TAG, NULL, tag_start, parse },
    { 0, NULL, NULL, tag_end, parse }
};

static state_r tag[] = {
    { 0, SPACE, NULL, tag, parse },
    { 0, OPEN_TAG, NULL, tag, parse },
    { 0, SLASH, NULL, tag, parse },

    { 6, "script", SPACE_CLOSE, NULL, freeze_tag },
    { 6, "object", SPACE_CLOSE, NULL, freeze_tag },
    { 6, "iframe", SPACE_CLOSE, NULL, freeze_tag },
    { 5, "style", SPACE_CLOSE, NULL, freeze_tag },
    { 5, "layer", SPACE_CLOSE, NULL, freeze_tag },
    { 6, "ilayer", SPACE_CLOSE, NULL, freeze_tag },
    { 6, "applet", SPACE_CLOSE, NULL, freeze_tag },
    { 5, "embed", SPACE_CLOSE, NULL, freeze_tag },

    { 0, NULL, NULL, param, parse }
};

static state_r param[] = {
    { 0, CLOSE_TAG, NULL, tag_start, parse },
    { 0, SPACE, NULL, param, parse },
    { 0, EQUAL, NULL, value, parse },
    { 2, "on", NULL, active_param, parse },
    { 5, "style", SPACE_EQ, NULL, freeze_param },
    { 0, NULL, NULL, param, parse }
};

static state_r active_param[] = {
    { 4, "Blur", SPACE_EQ, NULL, freeze_param },
    { 6, "Change", SPACE_EQ, NULL, freeze_param },
    { 5, "Focus", SPACE_EQ, NULL, freeze_param },
    { 6, "Select", SPACE_EQ, NULL, freeze_param },

    { 6, "Submit", SPACE_EQ, NULL, freeze_param },
    { 5, "Reset", SPACE_EQ, NULL, freeze_param },

    { 5, "Click", SPACE_EQ, NULL, freeze_param },
    { 8, "DblClick", SPACE_EQ, NULL, freeze_param },
    { 7, "MouseUp", SPACE_EQ, NULL, freeze_param },
    { 8, "MouseOut", SPACE_EQ, NULL, freeze_param },
    { 9, "MouseOver", SPACE_EQ, NULL, freeze_param },
    { 9, "MouseDown", SPACE_EQ, NULL, freeze_param },
    { 9, "MouseMove", SPACE_EQ, NULL, freeze_param },

    { 5, "KeyUp", SPACE_EQ, NULL, freeze_param },
    { 7, "KeyDown", SPACE_EQ, NULL, freeze_param },
    { 8, "KeyPress", SPACE_EQ, NULL, freeze_param },

    { 4, "Load", SPACE_EQ, NULL, freeze_param },
    { 6, "Unload", SPACE_EQ, NULL, freeze_param },
    { 5, "Abort", SPACE_EQ, NULL, freeze_param },
    { 5, "Error", SPACE_EQ, NULL, freeze_param },

    { 0, NULL, NULL, param, parse }
};

static state_r value[] = {
    { 0, CLOSE_TAG, NULL, tag_start, parse },
    { 0, SPACE, NULL, value, parse },
    { 0, QUOTE, NULL, value, parse },
    { 0, APOST, NULL, value, parse },

    { 11, "javascript:", NULL, NULL, freeze_scheme },
    { 11, "livescript:", NULL, NULL, freeze_scheme },
    { 9, "behavior:", NULL, NULL, freeze_scheme },
    { 9, "vbscript:", NULL, NULL, freeze_scheme },
    { 6, "about:", NULL, NULL, freeze_scheme },
    { 6, "mocha:", NULL, NULL, freeze_scheme },

    { 0, NULL, NULL, param, parse }
};


static state_r *state;
static void (*func)();

static unsigned char *buf;
static int n;
static int found;
static int max_len;
static int real_max_len;

static void freeze_add_module(module *m)
{
    if (m != &freeze_module)
        return;

    ap_hook_register("mod_accel::cont_tail",
                     freeze_content_tail_hook, AP_HOOK_NOCTX);
    ap_hook_register("mod_accel::range", freeze_range_hook, AP_HOOK_NOCTX);
    ap_hook_register("mod_accel::content", freeze_content_hook, AP_HOOK_NOCTX);
}

static void freeze_remove_module(module *m)
{
    if (m != &freeze_module)
        return;

    ap_hook_unregister("mod_accel::cont_tail", freeze_content_tail_hook);
    ap_hook_unregister("mod_accel::range", freeze_range_hook);
    ap_hook_unregister("mod_accel::content", freeze_content_hook);
}

static void freeze_init(server_rec *s, pool *p)
{
    if (ap_find_linked_module("mod_accel.c") == NULL) {
        ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ALERT, s,
                     "freeze: mod_accel is required");
        exit(1);
    }
}

static void freeze_content_tail_hook(server_rec *s, int *tail_len)
{
    if (ap_find_linked_module("mod_freeze.c") == NULL) {
        ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ALERT, s,
                     "freeze: add AddModule mod_freeze.c after "
                     "ClearModuleList directive");
        exit(1);
    }

    if (*tail_len < MAX_LEN)
        *tail_len = MAX_LEN;
}
 
static int freeze_range_hook(request_rec *r, header_rec *h)
{
    const char *ct;

    freeze_conf *conf = (freeze_conf *)
                   ap_get_module_config(r->per_dir_config, &freeze_module);

    /* there's no need to check CNT_FREEZE_(CHECKED|FOUND) because
       because CNT_RANGE_(CHECKED|DISABLED) checked in accel_handler() */

    if (conf->enable > 0
        && (ct = ap_table_get(h->headers_out, "Content-Type"))
        && strncasecmp(ct, "text/html", 9) == 0)
    {
        return DECLINED;
    }

    return OK;
}

static void freeze_content_hook(request_rec *r, accel_cnt *cnt)
{
    const char *ct;

    freeze_conf *conf = (freeze_conf *)
                     ap_get_module_config(r->per_dir_config, &freeze_module);

    if (cnt->state & CNT_FIRST)
        found = 0;

    if (conf->enable > 0
        && (cnt->in_flags & (CNT_FREEZE_CHECKED | CNT_FREEZE_FOUND))
                          != CNT_FREEZE_CHECKED
        /* Non-cached responses should not contain ranges */
        && (!ap_table_get(r->headers_out, "Content-Range"))
        && (ct = ap_table_get(r->headers_out, "Content-Type"))
        && strncasecmp(ct, "text/html", 9) == 0)
    {
        if (cnt->state & CNT_FIRST) {
            n = 0;

            if (conf->start_len == 0) {
                state = tag_start;

            } else {
                state = tag_begin;
                tag_begin[0].name = conf->start;
                tag_begin[0].len = conf->start_len;
            }

            func = parse;
        }

        buf = cnt->buf;
        max_len = cnt->len;

        if (cnt->state & CNT_TAIL) {
            real_max_len = max_len + MAX_LEN;

        } else {
            real_max_len = max_len;
        }

        while (n < max_len)
            (*func)();

        n -= max_len;
    }

    if (found)
        cnt->out_flags |= CNT_FREEZE_FOUND | CNT_RANGE_DISABLED;

    cnt->out_flags |= CNT_FREEZE_CHECKED | CNT_RANGE_CHECKED;
}


static void parse()
{
    int i, j;

    for (i = 0; state[i].name != NULL; i++) {

        if ((state[i].name == SPACE
                && (buf[n] == ' ' || buf[n] == '\n'
                    || buf[n] == '\r' || buf[n] == '\t'))
            || (state[i].name == OPEN_TAG
                && buf[n] == '<')
            || (state[i].name == EQUAL
                && buf[n] == '=')
            || (state[i].name == QUOTE
                && buf[n] == '"')
            || (state[i].name == APOST
                && buf[n] == '\'')
            || (state[i].name == SLASH
                && buf[n] == '/')
            || (state[i].name == CLOSE_TAG
                && buf[n] == '>')
           )
        {
            n++;
            func = state[i].f;
            state = state[i].s;

            return;
        }

        if (state[i].len == 0 || n + state[i].len >= real_max_len)
            continue;

        if (strncasecmp(&buf[n], state[i].name, state[i].len) == 0) {
            if (state[i].delim == NULL) {
                n += state[i].len;
                func = state[i].f;
                state = state[i].s;

                return;

            } else {
                for (j = 0; state[i].delim[j] != 0; j++) {
                    if (buf[n + state[i].len] == state[i].delim[j]) {
                        n += state[i].len;
                        func = state[i].f;
                        state = state[i].s;

                        return;
                    }
                }
            }
        }
    }

    n++;
    func = state[i].f;
    state = state[i].s;
}

static void freeze_tag()
{
    buf[n - 1] = 'X';

    found = 1;

    func = parse;
    state = tag_end;
}

static void freeze_param()
{
    buf[n - 1] = 'X';

    found = 1;

    func = parse;
    state = active_param;
}

static void freeze_scheme()
{
    buf[n - 2] = 'X';

    found = 1;

    func = parse;
    state = param;
}


static void *freeze_create_dir_config(pool *p, char *dir)
{
    freeze_conf *conf = ap_pcalloc(p, sizeof(freeze_conf));

    conf->enable = UNSET;
    conf->start = NULL;
    conf->start_len = 0;

    return (void *)conf;
}

static void *freeze_merge_dir_config(pool *p, void *base, void *new)
{
    freeze_conf *merged = ap_pcalloc(p, sizeof(freeze_conf));
    freeze_conf *parent = (freeze_conf *) base;
    freeze_conf *child = (freeze_conf *) new;

    merged->enable = child->enable == UNSET ? parent->enable : child->enable;
    merged->start = child->start == NULL ? parent->start : child->start;
    merged->start_len = child->start == NULL ? parent->start_len:
                                               child->start_len;

    return (void *)merged;
}

static const char *freeze_start(cmd_parms *cmd, void *mconfig, char *start)
{
    freeze_conf *conf = (freeze_conf *)mconfig;

    conf->start = start;
    conf->start_len = strlen(start);

    return NULL;
}


static const command_rec freeze_cmds[] =
{
    {"FreezeStart", freeze_start, NULL, RSRC_CONF | ACCESS_CONF, TAKE1,
     "tag or phrase to start freeze tags and parameters"},

    {"FreezeTags", ap_set_flag_slot,
     (void *)XtOffsetOf(freeze_conf, enable), RSRC_CONF | ACCESS_CONF, FLAG,
     "turn on/off too active tags and parameters freeze"},

    {NULL}
};


module freeze_module =
{
    STANDARD_MODULE_STUFF,
    freeze_init,                      /* module initializer */
    freeze_create_dir_config,         /* per-directory config creator */
    freeze_merge_dir_config,          /* dir config merger */
    NULL,                             /* server config creator */
    NULL,                             /* server config merger */
    freeze_cmds,                      /* command table */
    NULL,                             /* [9] list of handlers */
    NULL,                             /* [2] filename-to-URI translation */
    NULL,                             /* [5] check/validate user_id */
    NULL,                             /* [6] check user_id is valid *here* */
    NULL,                             /* [4] check access by host address */
    NULL,                             /* [7] MIME type checker/setter */
    NULL,                             /* [8] fixups */
    NULL,                             /* [10] logger */
    NULL,                             /* [3] header parser */
    NULL,                             /* process initializer */
    NULL,                             /* process exit/cleanup */
    NULL,                             /* [1] post read_request handling */

    /* EAPI */
    freeze_add_module,                /* after modules was added to core */
    freeze_remove_module,             /* before module is removed from core */
};
