/*
 * Copyright (C) 2001 Igor Sysoev.
 */

/*
MODULE-DEFINITION-START
Name: quoted_module
ConfigStart
ConfigEnd
MODULE-DEFINITION-END
*/

#ifndef EAPI
#error "mod_quoted requires Extended API (EAPI)"
#endif

#ifndef RUSSIAN_APACHE
#error "mod_quoted requires Russian Apache (mod_charset)"
#endif

#include "httpd.h"
#include "http_config.h"
#include "http_log.h"

#include "../accel/ap_accel.h"

#define UNSET   -1
#define MAX_LEN 5       /* "href=" */

#define SPACE           (char *)-1
#define OPEN_TAG        (char *)-2
#define CLOSE_TAG       (char *)-3

#define SPACE_DELIM     " \n\r\t"
#define SPACE_EQ        "= \n\r\t"


typedef struct {
    int           enable;
} quoted_conf;

typedef struct state_s {
    int                   len;
    char                 *name;
    char                 *delim;
    struct state_s       *s;
    void                (*f)();
} state_r;

module quoted_module;

static void parse ();
static void recode_quoted ();
static void quoted_content_tail_hook(server_rec *s, int *tail_len);
static int  quoted_range_hook(request_rec *r, header_rec *h);
static void quoted_content_hook(request_rec *r, accel_cnt *cnt);

static state_r tag[];
static state_r a_space_tag[];
static state_r a_tag[];
static state_r a_href_tag[];


static state_r tag_start[] =   {
                                 { 0, OPEN_TAG, NULL, tag, parse },
                                 { 0, NULL, NULL, tag_start, parse }
                               };

static state_r tag_end[] =     {
                                 { 0, CLOSE_TAG, NULL, tag_start, parse },
                                 { 0, NULL, NULL, tag_end, parse }
                               };

static state_r tag[] =         {
                                 { 0, SPACE, NULL, tag, parse },
                                 { 1, "a", SPACE_DELIM, a_space_tag, parse },
                                 { 0, NULL, NULL, tag_end, parse }
                               };

static state_r a_space_tag[] = {
                                 { 0, SPACE, NULL, a_space_tag, parse },
                                 { 0, CLOSE_TAG, NULL, tag_start, parse },
                                 { 4, "href", SPACE_EQ, a_href_tag, parse },
                                 { 0, NULL, NULL, a_tag, parse }
                               };

static state_r a_tag[] =       {
                                 { 0, SPACE, NULL, a_space_tag, parse },
                                 { 0, CLOSE_TAG, NULL, tag_start, parse },
                                 { 0, NULL, NULL, a_tag, parse }
                               };

static state_r a_href_tag[] =  {
                                 { 0, SPACE, NULL, a_href_tag, parse },
                                 { 1, "=", NULL, NULL, recode_quoted },
                                 { 0, NULL, NULL, tag_end, parse }
                               };



static state_r *state;
static void (*func)();

static request_rec *req;
static unsigned char *buf;
static int n;
static int found;
static int max_len;
static int quoted_max_len;
static int real_max_len;

static void quoted_add_module(module *m)
{
    if (m != &quoted_module)
        return;

    ap_hook_register("mod_accel::cont_tail",
                     quoted_content_tail_hook, AP_HOOK_NOCTX);
    ap_hook_register("mod_accel::range", quoted_range_hook, AP_HOOK_NOCTX);
    ap_hook_register("mod_accel::content", quoted_content_hook, AP_HOOK_NOCTX);
}

static void quoted_remove_module(module *m)
{
    if (m != &quoted_module)
        return;

    ap_hook_unregister("mod_accel::cont_tail", quoted_content_tail_hook);
    ap_hook_unregister("mod_accel::range", quoted_range_hook);
    ap_hook_unregister("mod_accel::content", quoted_content_hook);
}

static void quoted_init(server_rec *s, pool *p)
{
    if (ap_find_linked_module("mod_accel.c") == NULL) {
        ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ALERT, s,
                     "quoted: mod_accel is required");
        exit(1);
    }
}

static void quoted_content_tail_hook(server_rec *s, int *tail_len)
{
    if (ap_find_linked_module("mod_quoted.c") == NULL) {
        ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ALERT, s,
                     "quoted: add AddModule mod_quoted.c after "
                     "ClearModuleList directive");
        exit(1);
    }

    if (*tail_len < MAX_LEN)
        *tail_len = MAX_LEN;
}
 
static int quoted_range_hook(request_rec *r, header_rec *h)
{
    const char *ct;

    quoted_conf *conf = (quoted_conf *)
                   ap_get_module_config(r->per_dir_config, &quoted_module);

    /* there's no need to check CNT_QUOTED_(CHECKED|FOUND) because
       because CNT_RANGE_(CHECKED|DISABLED) checked in accel_handler() */

    if (conf->enable > 0
        && (ct = ap_table_get(h->headers_out, "Content-Type"))
        && strncasecmp(ct, "text/html", 9) == 0)
    {
        return DECLINED;
    }

    return OK;
}

static void quoted_content_hook(request_rec *r, accel_cnt *cnt)
{
    const char *ct;

    quoted_conf *conf = (quoted_conf *)
                     ap_get_module_config(r->per_dir_config, &quoted_module);

    if (cnt->state & CNT_FIRST)
        found = 0;

    if (conf->enable > 0
        && (cnt->in_flags & (CNT_QUOTED_CHECKED | CNT_QUOTED_FOUND))
                          != CNT_QUOTED_CHECKED
        /* Non-cached responses should not contain ranges */
        && (!ap_table_get(r->headers_out, "Content-Range"))
        && (ct = ap_table_get(r->headers_out, "Content-Type"))
        && strncasecmp(ct, "text/html", 9) == 0
        && ra_charset_ok(r))
    {
        if (cnt->state & CNT_FIRST) {
            req = r;
            n = 0;
            state = tag_start;
            func = parse;
        }

        buf = cnt->buf;
        max_len = cnt->len;

        if (cnt->state & CNT_TAIL) {
            quoted_max_len = max_len;
            real_max_len = max_len + MAX_LEN;

        } else {
            quoted_max_len = max_len - 2;
            real_max_len = max_len;
        }

        while (n < max_len)
            (*func)();

        n -= max_len;
    }

    if (found)
        cnt->out_flags |= CNT_QUOTED_FOUND | CNT_RANGE_DISABLED;

    cnt->out_flags |= CNT_QUOTED_CHECKED | CNT_RANGE_CHECKED;
}


static void parse()
{
    int i, j;

    for (i = 0; state[i].name != NULL; i++) {

        if ((state[i].name == SPACE
                && (buf[n] == ' ' || buf[n] == '\n'
                    || buf[n] == '\r' || buf[n] == '\t'))
            || (state[i].name == OPEN_TAG
                && buf[n] == '<')
            || (state[i].name == CLOSE_TAG
                && buf[n] == '>')
           )
        {
            n++;
            func = state[i].f;
            state = state[i].s;

            return;
        }

        if (state[i].len == 0 || n + state[i].len >= real_max_len)
            continue;

        if (strncasecmp(&buf[n], state[i].name, state[i].len) == 0) {
            if (state[i].delim == NULL) {
                n += state[i].len;
                func = state[i].f;
                state = state[i].s;

                return;

            } else {
                for (j = 0; state[i].delim[j] != 0; j++) {
                    if (buf[n + state[i].len] == state[i].delim[j]) {
                        n += state[i].len;
                        func = state[i].f;
                        state = state[i].s;

                        return;
                    }
                }
            }
        }
    }

    n++;
    func = state[i].f;
    state = state[i].s;
}

static void recode_quoted()
{
    int start, end;

    start = -1;

    while (n < max_len
           && (buf[n] == ' ' || buf[n] == '\n'
               || buf[n] == '\r' || buf[n] == '\t'))
        n++;

    for (; n < quoted_max_len && buf[n] != ' ' && buf[n] != '>'
                              && buf[n] != '\n' && buf[n] != '\r'
                              && buf[n] != '\t';
         n++) {

        if (buf[n] > 0x7f && start >= 0) {
            ra_in_place_convert_by_table_esc(&buf[start], end - start,
                                             req->ra_codep->cp_otabl_p);
            found++;
            start = -1;
            continue;
        }

        if (buf[n] == '%' && buf[n + 1] > '7') {
            if (start < 0)
                start = n;
            end = n + 3;
            n += 2;
        }
    }

    if (start >= 0) {
        ra_in_place_convert_by_table_esc(&buf[start], end - start,
                                         req->ra_codep->cp_otabl_p);
        found++;
    }

    func = parse;
    state = tag_end;
}


static void *quoted_create_dir_config(pool *p, char *dir)
{
    quoted_conf *conf = ap_pcalloc(p, sizeof(quoted_conf));

    conf->enable = UNSET;

    return (void *)conf;
}

static void *quoted_merge_dir_config(pool *p, void *base, void *new)
{
    quoted_conf *merged = ap_pcalloc(p, sizeof(quoted_conf));
    quoted_conf *parent = (quoted_conf *) base;
    quoted_conf *child = (quoted_conf *) new;

    merged->enable = child->enable == UNSET ? parent->enable : child->enable;

    return (void *)merged;
}


static const command_rec quoted_cmds[] =
{
    {"RecodeQuoted", ap_set_flag_slot,
     (void *)XtOffsetOf(quoted_conf, enable), RSRC_CONF | ACCESS_CONF, FLAG,
     "Turn on/off quoted recode"},

    {NULL}
};


module quoted_module =
{
    STANDARD_MODULE_STUFF,
    quoted_init,                      /* module initializer */
    quoted_create_dir_config,         /* per-directory config creator */
    quoted_merge_dir_config,          /* dir config merger */
    NULL,                             /* server config creator */
    NULL,                             /* server config merger */
    quoted_cmds,                      /* command table */
    NULL,                             /* [9] list of handlers */
    NULL,                             /* [2] filename-to-URI translation */
    NULL,                             /* [5] check/validate user_id */
    NULL,                             /* [6] check user_id is valid *here* */
    NULL,                             /* [4] check access by host address */
    NULL,                             /* [7] MIME type checker/setter */
    NULL,                             /* [8] fixups */
    NULL,                             /* [10] logger */
    NULL,                             /* [3] header parser */
    NULL,                             /* process initializer */
    NULL,                             /* process exit/cleanup */
    NULL,                             /* [1] post read_request handling */

    /* EAPI */
    quoted_add_module,                /* after modules was added to core */
    quoted_remove_module,             /* before module is removed from core */
};
