/*
 * Copyright (C) 2001 Igor Sysoev.
 */

/*
   RandomBanner  random=rb  10

   <a href="http://host/path0?place=1&random=rb00000">
   <img src="http://host/path1?place=1&random=rb00000">
   </a>

   <a href="http://host/path0?place=1&key=1234">
   <img src="http://host/path1?place=1&key=1234&random=rb00000">
   </a>

   RandomBanner  on|off
*/

/*
MODULE-DEFINITION-START
Name: randban_module
ConfigStart
ConfigEnd
MODULE-DEFINITION-END
*/

#ifndef EAPI
#error "mod_randban requires Extended API (EAPI)"
#endif

#include "httpd.h"
#include "http_config.h"
#include "http_log.h"

#include "../accel/ap_accel.h"

#define UNSET   -1

typedef struct {
    char         *match;
    int           len;
    int           max;
} randban_match;

typedef struct {
    int           enable;
    array_header *match;
} randban_conf;


module randban_module;

static int  randban_matched(array_header *match, char *buf, int len, int max,
                            request_rec *r);
static void randban_content_tail_hook(server_rec *s, int *tail_len);
static int  randban_ims_hook(request_rec *r, header_rec *h);
static int  randban_range_hook(request_rec *r, header_rec *h);
static void randban_content_hook(request_rec *r, accel_cnt *cnt);


static int max_len = 0;

/* non thread-safe */
static int last_len = 0;
static char *last_value;
static char *last_random;
static char fmt[] = "%0######d"; /* 1 million match length */

static void randban_add_module(module *m)
{
    if (m != &randban_module)
        return;

    ap_hook_register("mod_accel::cont_tail",
                     randban_content_tail_hook, AP_HOOK_NOCTX);
    ap_hook_register("mod_accel::ims", randban_ims_hook, AP_HOOK_NOCTX);
    ap_hook_register("mod_accel::range", randban_range_hook, AP_HOOK_NOCTX);
    ap_hook_register("mod_accel::content", randban_content_hook, AP_HOOK_NOCTX);
}

static void randban_remove_module(module *m)
{
    if (m != &randban_module)
        return;

    ap_hook_unregister("mod_accel::cont_tail", randban_content_tail_hook);
    ap_hook_unregister("mod_accel::ims", randban_ims_hook);
    ap_hook_unregister("mod_accel::range", randban_range_hook);
    ap_hook_unregister("mod_accel::content", randban_content_hook);
}

static void randban_init(server_rec *s, pool *p)
{
    if (ap_find_linked_module("mod_accel.c") == NULL) {
        ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ALERT, s,
                     "randban: mod_accel is required");
        exit(1);
    }

    last_value = ap_palloc(p, max_len);
    last_random = ap_palloc(p, max_len);
}

static void randban_child_init(server_rec *s, pool *p)
{
    srandom(getpid());
}

static void randban_content_tail_hook(server_rec *s, int *tail_len)
{
    if (ap_find_linked_module("mod_randban.c") == NULL) {
        ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ALERT, s,
                     "randban: add AddModule mod_randban.c after "
                     "ClearModuleList directive");
        exit(1);
    }

    if (*tail_len < max_len)
        *tail_len = max_len;
}

static int randban_ims_hook(request_rec *r, header_rec *h)
{
    const char *ct;

    randban_conf *conf = (randban_conf *)
                   ap_get_module_config(r->per_dir_config, &randban_module);

    if (conf->enable == UNSET)
        conf->enable = conf->match->nelts ? 1 : 0;

    /* there's no need to check CNT_RANDBAN_(CHECKED|FOUND) because
       because CNT_NM_(CHECKED|DISABLED) checked in accel_check_ims() */

    if (conf->enable > 0
        && (ct = ap_table_get(h->headers_out, "Content-Type"))
        && strncasecmp(ct, "text/html", 9) == 0)
        return DECLINED;

    return OK;
}

static int randban_range_hook(request_rec *r, header_rec *h)
{
    const char *ct;

    randban_conf *conf = (randban_conf *)
                   ap_get_module_config(r->per_dir_config, &randban_module);

    if (conf->enable == UNSET)
        conf->enable = conf->match->nelts ? 1 : 0;

    /* there's no need to check CNT_RANDBAN_(CHECKED|FOUND) because
       because CNT_RANGE_(CHECKED|DISABLED) checked in accel_handler() */

    if (conf->enable > 0
        && (ct = ap_table_get(h->headers_out, "Content-Type"))
        && strncasecmp(ct, "text/html", 9) == 0)
        return DECLINED;

    return OK;
}

static void randban_content_hook(request_rec *r, accel_cnt *cnt)
{
    const char *ct;

    randban_conf *conf = (randban_conf *)
                     ap_get_module_config(r->per_dir_config, &randban_module);

    if (cnt->state & CNT_FIRST)
        last_len = 0;

    if (conf->enable == UNSET)
        conf->enable = conf->match->nelts ? 1 : 0;

    if (conf->enable > 0
        && (cnt->in_flags & (CNT_RANDBAN_CHECKED | CNT_RANDBAN_FOUND))
                          != CNT_RANDBAN_CHECKED
        && (ct = ap_table_get(r->headers_out, "Content-Type"))
        && strncasecmp(ct, "text/html", 9) == 0)
    {
        if (randban_matched(conf->match, cnt->buf, cnt->len,
                            cnt->state & CNT_TAIL ? cnt->len + max_len:
                                                    cnt->len,
                            r))
        {
            cnt->out_flags |= CNT_RANDBAN_FOUND
                              | CNT_NM_DISABLED
                              | CNT_RANGE_DISABLED;
        }
    }

    cnt->out_flags |= CNT_RANDBAN_CHECKED | CNT_NM_CHECKED | CNT_RANGE_CHECKED;
}

static int randban_matched(array_header *match, char *buf, int len, int max,
                           request_rec *r)
{
    int            i, m, s, e, found;
    randban_match *elt;

    if (!match)
        return 0;

    found = 0;
    elt = (randban_match *)match->elts;

    for (i = 0; i < len; i++) {

        for (m = 0; m < match->nelts; m++) {

            if (buf[i] == elt[m].match[0]) {
                for (s = 1; i + s < max && s < elt[m].len; s++)
                    if (buf[i + s] != elt[m].match[s])
                        break;

                if (s < elt[m].len)
                    continue;

                s += i;
                for (e = s;
                     e < max && e < s + elt[m].max
                             && buf[e] >= '0' && buf[e] <= '9';
                     e++);

                if (e > s) {
                    found++;

                    if (e - s == last_len
                        && strncmp(&buf[s], last_value, last_len) == 0) {
                        memcpy(&buf[s], last_random, last_len);

                    } else {
                        int x, max_value = 1;
                        for (x = 0; x < e - s; x++)
                            max_value *= 10;

                        last_len = e - s;
                        ap_snprintf(fmt, sizeof(fmt), "%%0%dd", last_len),
                        ap_snprintf(last_random, last_len + 1, fmt,
                                    random() % max_value);
                        memcpy(last_value, &buf[s], last_len);
                        memcpy(&buf[s], last_random, last_len);
                    }

                    i = s - 1;
                    break;
                }
            }
        }
    }

    return found;
}


static void *randban_create_dir_config(pool *p, char *dir)
{
    randban_conf *conf = ap_pcalloc(p, sizeof(randban_conf));

    conf->match = ap_make_array(p, 10, sizeof(randban_match));
    conf->enable = UNSET;

    return (void *)conf;
}

static void *randban_merge_dir_config(pool *p, void *base, void *new)
{
    randban_conf *merged = ap_pcalloc(p, sizeof(randban_conf));
    randban_conf *parent = (randban_conf *) base;
    randban_conf *child = (randban_conf *) new;

    merged->match = ap_append_arrays(p, parent->match, child->match);

    merged->enable = child->enable != UNSET ?
                     child->enable : merged->match->nelts ?
                                     1 : parent->enable != UNSET ?
                                         parent->enable : 0;

    return (void *)merged;
}

static const char *add_random_banner(cmd_parms *cmd, void *dummy,
                                     char *one, char *two)
{
    int error = 0;
    randban_conf *conf = (randban_conf *) dummy;

    if (two) {
        randban_match *new = ap_push_array(conf->match);

        new->match = one;
        new->len = strlen(one);

        if ((new->max = atoi(two)) > 0) {
            if (new->len + new->max > max_len)
                max_len = new->len + new->max;

        } else {
            error = 1;
        }

    } else {

        if (strcasecmp(one, "on") == 0)
            conf->enable = 1;
        else if (strcasecmp(one, "off") == 0)
            conf->enable = 0;
        else
            error = 1;
    }

    if (error)
        return "error in RandomBanner directive";
    else
        return NULL;
}


static const command_rec randban_cmds[] =
{
    {"RandomBanner", add_random_banner, NULL, RSRC_CONF | ACCESS_CONF, TAKE12,
     "define matching string or turn on/off"},

    {NULL}
};


module randban_module =
{
    STANDARD_MODULE_STUFF,
    randban_init,                     /* module initializer */
    randban_create_dir_config,        /* per-directory config creator */
    randban_merge_dir_config,         /* dir config merger */
    NULL,                             /* server config creator */
    NULL,                             /* server config merger */
    randban_cmds,                     /* command table */
    NULL,                             /* [9] list of handlers */
    NULL,                             /* [2] filename-to-URI translation */
    NULL,                             /* [5] check/validate user_id */
    NULL,                             /* [6] check user_id is valid *here* */
    NULL,                             /* [4] check access by host address */
    NULL,                             /* [7] MIME type checker/setter */
    NULL,                             /* [8] fixups */
    NULL,                             /* [10] logger */
    NULL,                             /* [3] header parser */
    randban_child_init,               /* process initializer */
    NULL,                             /* process exit/cleanup */
    NULL,                             /* [1] post read_request handling */

    /* EAPI */
    randban_add_module,               /* after modules was added to core */
    randban_remove_module             /* before module is removed from core */
};
